# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for goal-based RL and empowerment/unsupervised RL.

See:
  VGCRL https://arxiv.org/abs/2106.01404
"""


import copy
import functools
from typing import Any, Callable, Dict, Optional, List, Tuple
from brax.v1.envs.env import Env
from brax.v1.envs.env import State
from brax.v1.experimental import normalization
from brax.v1.experimental.braxlines.common import dist_utils
from brax.v1.experimental.composer import composer
from brax.v1.experimental.composer import observers
from brax.training import networks
from brax.training.ppo import StepData
import jax
import jax.numpy as jnp
import tensorflow_probability.substrates.jax as tfp

tfd = tfp.distributions

DISC_PARAM_NAME = 'vgcrl_disc_params'


class Discriminator(object):
  """Container for p(z), q(z|o) in empowerment."""

  def __init__(self,
               env: Env,
               q_fn='indexing',
               q_fn_params=None,
               dist_p='Uniform',
               dist_p_params=None,
               dist_q='FixedSigma',
               dist_q_params=None,
               z_size=0,
               obs_indices: Optional[List[Any]] = None,
               obs_scale: Optional[List[float]] = None,
               nonnegative_reward: bool = True,
               logits_clip_range: float = 10.0,
               param_name: str = DISC_PARAM_NAME,
               normalize_obs: bool = False,
               spectral_norm: bool = True):
    assert obs_scale is not None
    self.z_size = z_size
    self.env_obs_size = env.observation_size
    self.model = None
    self.param_name = param_name
    self.q_fn_str = q_fn
    self.q_fn_params = q_fn_params or {}
    self.normalize_obs = normalize_obs
    self.obs_indices, self.obs_labels = observers.index_preprocess(
        obs_indices, env)
    self.indexed_obs_size = len(self.obs_indices)
    self.obs_scale = jnp.array(obs_scale or 1.) * jnp.ones(
        self.indexed_obs_size)
    self.normalize_fn = normalization.make_data_and_apply_fn(
        [self.env_obs_size + self.z_size], self.normalize_obs)[1]
    self.spectral_norm = spectral_norm
    self.logits_clip_range = logits_clip_range
    self.nonnegative_reward = nonnegative_reward
    self.ll_q_offset = 0.0

    # define dist_params_to_dist for q_z_o
    dist_q_params = copy.deepcopy(dist_q_params) or {}
    q_scale = dist_q_params.pop('scale', 1.)
    q_scale = jnp.array(q_scale) * jnp.ones(self.z_size)
    assert z_size
    if dist_q == 'FixedSigma':
      self.dist_q_fn = lambda x: tfd.MultivariateNormalDiag(x, q_scale)
      # if an environment terminates, it's useful to ensure reward is positive,
      #  this ensures that within 3*std the log likelihood is positive.
      if self.nonnegative_reward:
        self.ll_q_offset = -tfd.MultivariateNormalDiag(
            jnp.zeros(self.z_size),
            jnp.ones(self.z_size) * q_scale).log_prob(
                jnp.ones(self.z_size) * q_scale * 3)
    elif dist_q == 'Categorial':
      self.dist_q_fn = lambda x: dist_utils.clipped_onehot_categorical(
          logits=x, clip_range=logits_clip_range)
      if self.nonnegative_reward:
        self.ll_q_offset = self.logits_clip_range
    else:
      raise NotImplementedError(dist_q)
    assert not dist_q_params, f'unused dist_q_params: {dist_q_params}'

    # define dist for p_z
    dist_p_params = copy.deepcopy(dist_p_params) or {}
    if dist_p == 'Uniform':
      p_scale = dist_p_params.pop('scale', 1.)
      p_scale = jnp.array(p_scale) * jnp.ones(self.z_size)
      self.dist_p_fn = lambda: tfd.Uniform(low=-p_scale, high=p_scale)
    elif dist_p == 'UniformCategorial':
      self.dist_p_fn = lambda: tfd.OneHotCategorical(
          logits=jnp.zeros(self.z_size))
    elif dist_p == 'Deterministic':
      p_value = dist_p_params.pop('value')
      p_value = jnp.array(p_value) * jnp.ones(self.z_size)
      self.dist_p_fn = lambda: tfd.Deterministic(loc=p_value)
    else:
      raise NotImplementedError(dist_p)
    assert not dist_p_params, f'unused dist_p_params: {dist_p_params}'

    self.initialized = False

  def init_model(self, rng: jnp.ndarray = None):
    """Initialize parts with model parameters."""
    model_params = None
    # define observation_to_dist_params mapping for q_z_o
    q_fn, q_fn_params = self.q_fn_str, self.q_fn_params
    if q_fn == 'indexing':
      self.q_fn = lambda params, x: (self.index_obs(x),)
    elif q_fn == 'mlp':
      input_size = q_fn_params.get('input_size')
      output_size = q_fn_params.get('output_size')
      model = networks.make_model([32, 32, output_size],
                                  input_size,
                                  spectral_norm=self.spectral_norm)
      self.model = model
      if self.spectral_norm:
        rng1, rng2, rng3 = jax.random.split(rng, 3)
        model_params = model.init(rng1, rng2)
        self.q_fn = lambda params, x: (model.apply(
            params, x, rngs={'sing_vec': rng3}, mutable=['sing_vec'])[0],)
      else:
        model_params = model.init(rng)
        self.q_fn = lambda params, x: (model.apply(params, x),)
    elif q_fn == 'indexing_mlp':
      output_size = q_fn_params.get('output_size')
      model = networks.make_model([32, 32, output_size],
                                  len(self.obs_indices),
                                  spectral_norm=self.spectral_norm)
      self.model = model
      if self.spectral_norm:
        rng1, rng2, rng3 = jax.random.split(rng, 3)
        model_params = model.init(rng1, rng2)
        self.q_fn = lambda params, x: (model.apply(
            params,
            self.index_obs(x),
            rngs={'sing_vec': rng3},
            mutable=['sing_vec'])[0],)
      else:
        model_params = model.init(rng)
        self.q_fn = lambda params, x: (model.apply(params, self.index_obs(x)),)
    else:
      raise NotImplementedError(q_fn)
    self.initialized = True
    return {self.param_name: model_params} if model_params else {}

  def dist_q_z_o(self,
                 data: jnp.ndarray,
                 params: Dict[str, Dict[str, jnp.ndarray]] = None):
    assert self.initialized, 'init_model() must be called'
    param = params.get(self.param_name, {})
    dist_params = self.q_fn(param, data)
    return self.dist_q_fn(*dist_params)

  def dist_p_z(self):
    return self.dist_p_fn()

  def sample_p_z(self, batch_size: int, rng: jnp.ndarray):
    """Sample from p(z)."""
    dist_p = self.dist_p_fn()
    if batch_size:
      return dist_p.sample(seed=rng, sample_shape=(batch_size,))
    else:
      return dist_p.sample(seed=rng, sample_shape=())

  def zero_z(self, batch_size: int):
    if batch_size:
      return jnp.zeros((batch_size, self.z_size))
    else:
      return jnp.zeros(self.z_size)

  def ll_q_z_o(self,
               z: jnp.ndarray,
               data: jnp.ndarray,
               params: Dict[str, Dict[str, jnp.ndarray]] = None,
               add_offset: bool = False):
    dist = self.dist_q_z_o(data, params=params)
    ll = dist.log_prob(z)
    if add_offset:
      ll += self.ll_q_offset
    return ll

  def split_obs(self, obs: jnp.ndarray):
    env_obs = obs[..., :-self.z_size]
    z = obs[..., -self.z_size:]
    return env_obs, z

  def index_obs(self, obs: jnp.ndarray):
    if self.obs_indices:
      return obs[..., self.obs_indices]
    return obs

  def unindex_obs(self, indexed_obs: jnp.ndarray):
    if self.obs_indices:
      obs = jnp.zeros(indexed_obs.shape[:-1] + (self.env_obs_size,))
      obs = obs.at[..., self.obs_indices].add(indexed_obs)
      return obs
    return indexed_obs

  def concat_obs(self, obs: jnp.ndarray, z: jnp.ndarray):
    new_obs = jnp.concatenate([obs, z], axis=-1)
    return new_obs

  def disc_loss_fn(self, data: StepData, udata: StepData, rng: jnp.ndarray,
                   params: Dict[str, Dict[str, jnp.ndarray]]):
    return disc_loss_fn(
        data, udata, rng, params, disc=self, normalize_obs=self.normalize_obs)


class ParameterizeWrapper(Env):
  """A wrapper that parameterizes Brax Env."""

  def __init__(self,
               environment: Env,
               disc: Discriminator,
               obs_norm_reward_multiplier: float = 0.0,
               env_reward_multiplier: float = 0.0):
    self._environment = environment
    self.action_repeat = self._environment.action_repeat
    if hasattr(self._environment, 'batch_size'):
      self.batch_size = self._environment.batch_size
    else:
      self.batch_size = None
    self.sys = self._environment.sys
    self.disc = disc
    self.z_size = disc.z_size
    self.env_obs_size = self._environment.observation_size
    self.env_reward_multiplier = env_reward_multiplier
    self.obs_norm_reward_multiplier = obs_norm_reward_multiplier

  def concat(
      self,
      state: State,
      z: jnp.ndarray,
      normalizer_params: Dict[str, jnp.ndarray] = None,
      params: Dict[str, Dict[str, jnp.ndarray]] = None,
      replace_reward: bool = True,
  ) -> State:
    """Concatenate state with param and recompute reward."""
    new_obs = self.disc.concat_obs(state.obs, z)
    state = state.replace(obs=new_obs)
    if replace_reward:
      new_obs = self.disc.normalize_fn(normalizer_params, new_obs)
      env_obs, z = self.disc.split_obs(new_obs)
      new_reward = self.disc.ll_q_z_o(
          z, env_obs, params=params, add_offset=True)
      if self.obs_norm_reward_multiplier:
        new_reward += self.obs_norm_reward_multiplier * jnp.linalg.norm(
            self.disc.index_obs(env_obs), axis=-1)
      new_reward = jax.lax.stop_gradient(new_reward)
      state = state.replace(reward=new_reward +
                            self.env_reward_multiplier * state.reward)
    return state

  def reset(self, rng: jnp.ndarray, z: jnp.ndarray = None) -> State:
    """Resets the environment to an initial state."""
    state = self._environment.reset(rng)
    if z is None:
      z = self.disc.sample_p_z(self.batch_size, rng)
    else:
      assert z.shape[-1] == self.z_size, f'{z.shape}[-1] != {self.z_size}'
    return self.concat(state, z, replace_reward=False)

  def step(self,
           state: State,
           action: jnp.ndarray,
           normalizer_params: Dict[str, jnp.ndarray] = None,
           extra_params: Dict[str, Dict[str, jnp.ndarray]] = None) -> State:
    """Run one timestep of the environment's dynamics."""
    _, z = self.disc.split_obs(state.obs)
    state = self._environment.step(state, action)
    return self.concat(
        state, z, normalizer_params, extra_params, replace_reward=True)

  def step2(self, state: State, action: jnp.ndarray) -> State:
    """Run one timestep of the environment's dynamics."""
    _, z = self.disc.split_obs(state.obs)
    state = self._environment.step(state, action)
    return self.concat(state, z, replace_reward=False)


def disc_loss_fn(data: StepData,
                 udata: StepData,
                 rng: jnp.ndarray,
                 params: Dict[str, Dict[str, jnp.ndarray]],
                 disc: Discriminator,
                 normalize_obs: bool = False):
  """Discriminator loss function."""
  disc_loss = 0
  if disc and disc.model:
    d = data if normalize_obs else udata
    env_obs, z = disc.split_obs(d.obs)
    disc_loss = -jnp.mean(disc.ll_q_z_o(z, env_obs, params=params))
  return disc_loss, rng


def create(env_name: str, wrapper_params: Dict[str, Any], **kwargs) -> Env:
  """Creates an Env with a specified brax system."""
  env = composer.create(env_name=env_name, **kwargs)
  return ParameterizeWrapper(env, **wrapper_params)


def create_fn(env_name: str, wrapper_params: Dict[str, Any],
              **kwargs) -> Callable[..., Env]:
  """Returns a function that when called, creates an Env."""
  return functools.partial(create, env_name, wrapper_params, **kwargs)


def create_disc_fn(algo_name: str,
                   observation_size: int,
                   obs_indices: Tuple[Any] = None,
                   scale: float = 1.0,
                   diayn_num_skills: int = 8,
                   logits_clip_range: float = 5.0,
                   spectral_norm: bool = False,
                   **kwargs):
  """Create a standard discriminator."""
  disc_fn = {
      'fixed_gcrl':
          functools.partial(
              Discriminator,
              q_fn='indexing',
              z_size=len(obs_indices),
              obs_indices=obs_indices,
              dist_p='Deterministic',
              dist_p_params=dict(value=scale),
          ),
      'gcrl':
          functools.partial(
              Discriminator,
              q_fn='indexing',
              z_size=len(obs_indices),
              obs_indices=obs_indices,
              dist_p_params=dict(scale=scale / 2.0),
          ),
      'cdiayn':
          functools.partial(
              Discriminator,
              q_fn='indexing_mlp',
              z_size=len(obs_indices),
              obs_indices=obs_indices,
              q_fn_params=dict(output_size=len(obs_indices),),
              spectral_norm=spectral_norm,
          ),
      'diayn':
          functools.partial(
              Discriminator,
              q_fn='indexing_mlp',
              z_size=diayn_num_skills,
              obs_indices=obs_indices,
              q_fn_params=dict(output_size=diayn_num_skills,),
              dist_p='UniformCategorial',
              dist_q='Categorial',
              logits_clip_range=logits_clip_range,
              spectral_norm=spectral_norm,
          ),
      'diayn_full':
          functools.partial(
              Discriminator,
              q_fn='mlp',
              z_size=diayn_num_skills,
              obs_indices=obs_indices,
              q_fn_params=dict(
                  input_size=observation_size,
                  output_size=diayn_num_skills,
              ),
              dist_p='UniformCategorial',
              dist_q='Categorial',
              logits_clip_range=logits_clip_range,
              spectral_norm=spectral_norm,
          ),
  }.get(algo_name, None)
  assert disc_fn, f'invalid algo_name: {algo_name}'
  disc_fn = functools.partial(disc_fn, obs_scale=scale, **kwargs)
  return disc_fn



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Components.

In Braxlines Composer, each environment is built reusing
components defined in this folder. This faciliates reusability,
procedural generation, and randomizability of Brax environments.
"""
import copy
import importlib
from typing import Any, Dict

DEFAULT_REGISTER_COMPONENTS = ('ant', 'ground', 'halfcheetah', 'singleton',
                               'pro_ant', 'octopus', 'humanoid')

COMPONENT_MAPPING = {}


def list_components():
  return sorted(COMPONENT_MAPPING)


def exists(component: str):
  return component in COMPONENT_MAPPING


def register_component(component: str,
                       component_specs: Any = None,
                       load_path: str = None,
                       override: bool = False):
  """Register component library."""
  global COMPONENT_MAPPING
  if not override and component in COMPONENT_MAPPING:
    return COMPONENT_MAPPING[component]
  if component_specs is None:
    if not load_path:
      if '.' not in component:
        load_path = f'brax.v1.experimental.composer.components.{component}'
      else:
        load_path = component
    component_lib = importlib.import_module(load_path)
  COMPONENT_MAPPING[component] = component_lib
  return component_lib


def register_default_components():
  """Register all default components."""
  for component in DEFAULT_REGISTER_COMPONENTS:
    register_component(component)


def load_component(component: str,
                   component_specs: Dict[str, Any] = None,
                   component_params: Dict[str, Any] = None,
                   **override_specs) -> Dict[str, Any]:
  """Load component config information."""
  if isinstance(component, str):
    # if string, load a library under composer/components
    component_lib = register_component(
        component=component, component_specs=component_specs)
    specs_fn = component_lib.get_specs
  else:
    # else, assume it's a custom get_specs()
    specs_fn = component
  default_specs = specs_fn(**(component_params or {}))
  default_specs = copy.deepcopy(default_specs)
  default_specs.update(override_specs)
  return default_specs



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import, g-importing-member
"""Functions for smooth forward and inverse dynamics."""
from brax import fluid
from brax import math
from brax import scan
from brax.base import Motion, System, Transform
from brax.generalized.base import State
from brax.generalized.constraint import point_jacobian
import jax
from jax import numpy as jp


def transform_com(sys: System, state: State) -> State:
  """Transforms inertia, dof, and link velocity into center of mass frame.

  Args:
    sys: a brax system
    state: generalized state

  Returns:
    state: generalized state with com, cinr, cd, cdof, cdofd updated
  """
  x_i = state.x.vmap().do(sys.link.inertia.transform)
  root_fn = lambda i, p=sys.link_parents: i if p[i] < 0 else root_fn(p[i])
  root = jp.array([root_fn(i) for i in range(sys.num_links())])
  mass_xi = jax.vmap(jp.multiply)(sys.link.inertia.mass, x_i.pos)
  mass_xi_sum = jax.ops.segment_sum(mass_xi, root, sys.num_links())
  mass_sum = jax.ops.segment_sum(sys.link.inertia.mass, root, sys.num_links())
  root_com = jax.vmap(jp.divide)(mass_xi_sum[root], mass_sum[root])
  cinr = x_i.replace(pos=x_i.pos - root_com).vmap().do(sys.link.inertia)

  # motion dofs to global frame centered at subtree-CoM
  parent_idx = jp.array(
      [
          i if t == 'f' else p
          for i, (t, p) in enumerate(zip(sys.link_types, sys.link_parents))
      ]
  )
  parent = state.x.concatenate(Transform.zero(shape=(1,))).take(parent_idx)
  j = parent.vmap().do(sys.link.transform).vmap().do(sys.link.joint)

  # propagate motion through stacked joints
  def cdof_fn(typ, q, motion):
    if typ == 'f':
      return motion

    # create joint transforms and motions:
    # - rotation/velocity about axis for revolute (motion.ang)
    # - translation/velocity along axis for prismatic (motion.vel)
    rot_fn = lambda ang, q: math.normalize(math.quat_rot_axis(ang, q))[0]
    j = Transform.create(
        rot=jax.vmap(rot_fn)(motion.ang, q),
        pos=jax.vmap(jp.multiply)(motion.vel, q),
    )
    jd = motion

    # then group them by link, so each has num_dofs joints
    num_links, num_dofs = motion.ang.shape[0] // int(typ), int(typ)
    s = (num_links, num_dofs, -1)
    j_stack, jd_stack = j.reshape(s), jd.reshape(s)

    # accumulate jds as successive j transforms around joint
    j, jds = Transform.zero(shape=(num_links,)), []
    for i in range(0, num_dofs):
      jds.append(j.vmap().inv_do(jd_stack.take(i, axis=1)))
      j = j.vmap().do(j_stack.take(i, axis=1))

    # interleave jds back together to match joint stack order
    motion = jax.tree.map(lambda *x: jp.column_stack(x), *jds).reshape((-1, 3))

    return motion

  cdof = scan.link_types(sys, cdof_fn, 'qd', 'd', state.q, sys.dof.motion)
  ang = jax.vmap(math.rotate)(cdof.ang, j.take(sys.dof_link()).rot)
  cdof = cdof.replace(ang=ang)
  off = Transform.create(pos=root_com - j.pos)
  cdof = off.take(sys.dof_link()).vmap().do(cdof)
  cdof_qd = jax.vmap(lambda x, y: x * y)(cdof, state.qd)

  # forward scan down tree: accumulate link center of mass velocity
  def cd_fn(cd_parent, cdof_qd, dof_idx):
    if cd_parent is None:
      num_roots = len([p for p in sys.link_parents if p == -1])
      cd_parent = Motion.zero(shape=(num_roots,))

    # cd = cd[parent] + map-sum(cdof * qd)
    cd = cd_parent.index_sum(dof_idx, cdof_qd)

    return cd

  cd = scan.tree(sys, cd_fn, 'dd', cdof_qd, sys.dof_link(depth=True))

  # propagate cd through stacked joints to calculate cdofd
  def cdofd_fn(typ, cd, cdof, cdof_qd):
    if typ == 'f':
      cdof_qd = cdof_qd.reshape((-1, 6, 3))
      cd = jax.tree.map(lambda x: jp.sum(x[:, 0:3], axis=1), cdof_qd)
      cdofd = cd.vmap().vmap(in_axes=(None, 0)).cross(cdof.reshape((-1, 6, 3)))
      cdofd = jax.tree.map(lambda x: x.at[:, 0:3].set(jp.zeros(3)), cdofd)
      return cdofd.reshape((-1, 3))

    # group cdof_qd by link, so each has num_dofs joints
    num_dofs = int(typ)
    cdof_qd = cdof_qd.reshape((cd.ang.shape[0], num_dofs, -1))
    cds = [cd]

    # accumulate cd as successive cdof velocities
    for i in range(0, num_dofs - 1):
      cds.append(cds[-1] + cdof_qd.take(i, axis=1))

    # interleave cds back together to match joint stack order
    cd = jax.tree.map(lambda *x: jp.column_stack(x), *cds).reshape((-1, 3))
    cdofd = cd.vmap().cross(cdof)

    return cdofd

  cd_p = cd.concatenate(Motion.zero(shape=(1,))).take(parent_idx)
  cdofd = scan.link_types(sys, cdofd_fn, 'ldd', 'd', cd_p, cdof, cdof_qd)

  return state.replace(
      root_com=root_com, cinr=cinr, cd=cd, cdof=cdof, cdofd=cdofd
  )


def inverse(sys: System, state: State) -> jax.Array:
  """Calculates the system's forces given input motions.

  This function computes inverse dynamics using the Newton-Euler algorithm:

  https://scaron.info/robot-locomotion/recursive-newton-euler-algorithm.html

  Args:
    sys: a brax system
    state: generalized state

  Returns:
    tau: generalized forces resulting from joint positions and velocities
  """
  # forward scan over tree: accumulate link center of mass acceleration
  def cdd_fn(cdd_parent, cdofd, qd, dof_idx):
    if cdd_parent is None:
      num_roots = len([p for p in sys.link_parents if p == -1])
      cdd_parent = Motion.create(vel=-jp.tile(sys.gravity, (num_roots, 1)))

    # cdd = cdd[parent] + map-sum(cdofd * qd)
    cdd = cdd_parent.index_sum(dof_idx, jax.vmap(lambda x, y: x * y)(cdofd, qd))

    return cdd

  cdd = scan.tree(
      sys, cdd_fn, 'ddd', state.cdofd, state.qd, sys.dof_link(depth=True)
  )

  # cfrc_flat = cinr * cdd + cd x (cinr * cd)
  def frc(cinr, cdd, cd):
    return cinr.mul(cdd) + cd.cross(cinr.mul(cd))

  cfrc_flat = jax.vmap(frc)(state.cinr, cdd, state.cd)

  # backward scan up tree: accumulate link center of mass forces
  def cfrc_fn(cfrc_child, cfrc):
    if cfrc_child is not None:
      cfrc += cfrc_child
    return cfrc

  cfrc = scan.tree(sys, cfrc_fn, 'l', cfrc_flat, reverse=True)

  # tau = cdof * cfrc[dof_link]
  tau = jax.vmap(lambda x, y: x.dot(y))(state.cdof, cfrc.take(sys.dof_link()))

  return tau


def _passive(sys: System, state: State) -> jax.Array:
  """Calculates the system's passive forces given input motion and position."""
  def stiffness_fn(typ, q, dof):
    if typ in 'fb':
      return jp.zeros_like(dof.stiffness)
    return -q * dof.stiffness

  frc = scan.link_types(sys, stiffness_fn, 'qd', 'd', state.q, sys.dof)
  frc -= sys.dof.damping * state.qd

  if sys.enable_fluid:
    fluid_frc = fluid.force(
        sys,
        state.x,
        state.cd,
        sys.link.inertia.mass,
        sys.link.inertia.i,
        state.root_com,
    )
    link_idx = jp.arange(sys.num_links())
    x_i = state.x.vmap().do(sys.link.inertia.transform)
    jac_fn = jax.vmap(point_jacobian, in_axes=(None, None, None, 0, 0))
    jac = jac_fn(sys, state.root_com, state.cdof, x_i.pos, link_idx)
    frc += jax.vmap(lambda x, y: x.dot(y))(jac, fluid_frc).sum(axis=0)

  return frc


def forward(sys: System, state: State, tau: jax.Array) -> jax.Array:
  """Calculates resulting joint forces given input forces.

  This method builds and solves the linear system: M @ qdd = -C + tau

  where M is the joint space inertia matrix, "mass matrix" and
        C is the bias force calculated by inverse()

  Args:
    sys: a brax system
    state: generalized state
    tau: joint force input vector

  Returns:
    qfrc: joint force vector
  """
  qfrc_passive = _passive(sys, state)
  qfrc_bias = inverse(sys, state)
  qfrc = qfrc_passive - qfrc_bias + tau

  return qfrc



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Tests for generalized pipeline."""

from absl.testing import absltest
from absl.testing import parameterized
from brax import test_utils
from brax.generalized import pipeline
from brax.io import mjcf
import jax
from jax import numpy as jp
import numpy as np


class PipelineTest(parameterized.TestCase):

  @parameterized.parameters(
      ('ant.xml',),
      ('triple_pendulum.xml',),
      ('humanoid.xml',),
      ('half_cheetah.xml',),
      ('swimmer.xml',),
  )
  def test_forward(self, xml_file):
    """Test pipeline step."""
    sys = test_utils.load_fixture(xml_file)
    # crank up solver iterations just to demonstrate close match to mujoco
    sys = sys.replace(solver_iterations=500)
    for mj_prev, mj_next in test_utils.sample_mujoco_states(xml_file):
      state = jax.jit(pipeline.init)(sys, mj_prev.qpos, mj_prev.qvel)
      state = jax.jit(pipeline.step)(sys, state, jp.zeros(sys.act_size()))

      np.testing.assert_allclose(state.q, mj_next.qpos, atol=0.002)
      np.testing.assert_allclose(state.qd, mj_next.qvel, atol=0.5)


class GradientTest(absltest.TestCase):
  """Tests that gradients are not NaN."""

  def test_grad(self):
    """Tests that gradients are not NaN."""
    xml = """
    <mujoco>
      <worldbody>
        <body>
          <joint type="slide" axis="1 0 0" damping="1"/>
          <joint type="slide" axis="0 1 0" damping="1"/>
          <geom size="0.1" mass="1"/>
        </body>
      </worldbody>
    </mujoco>
    """
    sys = mjcf.loads(xml)
    init_state = jax.jit(pipeline.init)(
        sys, sys.init_q, jp.zeros(sys.qd_size())
    )

    def fn(xd):
      qd = jp.zeros(sys.qd_size()).at[0].set(xd)
      state = init_state.replace(qd=qd)
      for _ in range(10):
        state = jax.jit(pipeline.step)(sys, state, None)
      return state.qd[0]

    grad = jax.grad(fn)(-1.0)
    self.assertFalse(np.isnan(grad))


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import, g-importing-member
"""Base types for generalized pipeline."""

from brax import base
from brax.base import Inertia, Motion, Transform
from flax import struct
import jax
from jax import numpy as jp


@struct.dataclass
class State(base.State):
  """Dynamic state that changes after every step.

  Attributes:
    root_com: (num_links,) center of mass position of link root kinematic tree
    cinr: (num_links,) inertia in com frame
    cd: (num_links,) link velocities in com frame
    cdof: (qd_size,) dofs in com frame
    cdofd: (qd_size,) cdof velocity
    mass_mx: (qd_size, qd_size) mass matrix
    mass_mx_inv: (qd_size, qd_size) inverse mass matrix
    contact: calculated contacts
    con_jac: constraint jacobian
    con_diag: constraint A diagonal
    con_aref: constraint reference acceleration
    qf_smooth: (qd_size,) smooth dynamics force
    qf_constraint: (qd_size,) force from constraints (collision etc)
    qdd: (qd_size,) joint acceleration vector
  """

  # position/velocity based terms are updated at the end of each step:
  root_com: jax.Array
  cinr: Inertia
  cd: Motion
  cdof: Motion
  cdofd: Motion
  mass_mx: jax.Array
  mass_mx_inv: jax.Array
  con_jac: jax.Array
  con_diag: jax.Array
  con_aref: jax.Array
  # acceleration based terms are calculated using terms from the previous step:
  qf_smooth: jax.Array
  qf_constraint: jax.Array
  qdd: jax.Array

  @classmethod
  def init(
      cls, q: jax.Array, qd: jax.Array, x: Transform, xd: Motion
  ) -> 'State':
    """Returns an initial State given a brax system."""
    num_links = x.pos.shape[0]
    qd_size = qd.shape[0]
    return State(
        q=q,
        qd=qd,
        x=x,
        xd=xd,
        contact=None,
        root_com=jp.zeros(3),
        cinr=Inertia(
            Transform.zero((num_links,)),
            jp.zeros((num_links, 3, 3)),
            jp.zeros((num_links,)),
        ),
        cd=Motion.zero((num_links,)),
        cdof=Motion.zero((num_links,)),
        cdofd=Motion.zero((num_links,)),
        mass_mx=jp.zeros((qd_size, qd_size)),
        mass_mx_inv=jp.zeros((qd_size, qd_size)),
        con_jac=jp.zeros(()),
        con_diag=jp.zeros(()),
        con_aref=jp.zeros(()),
        qf_smooth=jp.zeros_like(qd),
        qf_constraint=jp.zeros_like(qd),
        qdd=jp.zeros_like(qd),
    )



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""General-purpose file interface to support different backing file APIs."""


import glob
import os
from typing import List


def Glob(pattern: str) -> List[str]:
  return glob.glob(pattern)


def Exists(pathname: str) -> bool:
  """Check if a path exists."""
  exist = None
  if exist is None:
    exist = os.path.exists(pathname)
  return exist


def MakeDirs(dirname: str):
  """Make directory if it doesn't exist."""
  exist_dir = False
  if not exist_dir:
    os.makedirs(dirname, exist_ok=True)


class File:
  """General purpose file resource."""

  def __init__(self, fileName: str, mode='r'):
    self.f = None
    if not self.f:
      self.f = open(fileName, mode)

  def __enter__(self):
    return self.f

  def __exit__(self, exc_type, exc_value, traceback):
    self.f.close()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for geometry."""

from absl.testing import absltest
from absl.testing import parameterized
from brax import math
import jax
from jax import numpy as jp
import numpy as np


def _get_rand_norm(seed: int):
  np.random.seed(seed)
  theta = np.random.random(1) * 2 * np.pi
  a = (np.random.random(1) - 0.5) * 2.0
  phi = np.arccos(a)
  x = np.sin(phi) * np.cos(theta)
  y = np.sin(phi) * np.sin(theta)
  z = np.cos(phi)
  return jp.array([x, y, z]).squeeze()


class MathTest(absltest.TestCase):

  def test_inv_approximate(self):
    # create a 4x4 matrix we know is invertible
    x = jax.random.normal(jax.random.PRNGKey(0), (4, 4))
    x = jp.eye(4) * 0.001 + x @ x.T

    x_inv = jp.linalg.inv(x)
    x_inv_approximate = math.inv_approximate(x, jp.zeros((4, 4)), num_iter=100)

    np.testing.assert_array_almost_equal(x_inv_approximate, x_inv)

  def test_from_to(self):
    v1 = jp.array([1.0, 0.0, 0.0])
    rot = math.from_to(v1, v1)
    np.testing.assert_array_almost_equal(v1, math.rotate(v1, rot))

    rot = math.from_to(v1, -v1)
    np.testing.assert_array_almost_equal(-v1, math.rotate(v1, rot))

    rot = math.from_to(-v1, v1)
    np.testing.assert_array_almost_equal(v1, math.rotate(-v1, rot))

    v1 = jp.array([0.0, 1.0, 0.0])
    rot = math.from_to(v1, -v1)
    np.testing.assert_array_almost_equal(-v1, math.rotate(v1, rot))

    v2 = jp.array([-0.5, 0.5, 0.0])
    v2 /= jp.linalg.norm(v2)
    rot = math.from_to(v1, v2)
    np.testing.assert_array_almost_equal(v2, math.rotate(v1, rot))


class OrthoganalsTest(parameterized.TestCase):
  """Tests the orthogonals function."""

  @parameterized.parameters(range(100))
  def test_orthogonals(self, i):
    a = _get_rand_norm(i)
    b, c = math.orthogonals(a)
    np.testing.assert_almost_equal(jp.linalg.norm(a), 1)
    np.testing.assert_almost_equal(jp.linalg.norm(b), 1)
    np.testing.assert_almost_equal(jp.linalg.norm(c), 1)
    self.assertAlmostEqual(np.abs(a.dot(b)), 0, 6)
    self.assertAlmostEqual(np.abs(b.dot(c)), 0, 6)
    self.assertAlmostEqual(np.abs(a.dot(c)), 0, 6)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions for integrating maximal coordinate dynamics."""
# pylint:disable=g-multiple-import
from typing import Tuple

from brax import math
from brax.base import Motion, System, Transform
import jax
from jax import numpy as jp


def integrate_xdv(sys: System, xd: Motion, xdv: Motion) -> Motion:
  """Updates velocity by applying delta-velocity.

  Args:
    sys: System to forward propagate
    xd: velocity
    xdv: delta-velocity

  Returns:
    xd: updated velocity
  """
  damp = Motion(vel=sys.vel_damping, ang=sys.ang_damping)
  xd = (
      jax.tree.map(lambda d, x: jp.exp(d * sys.opt.timestep) * x, damp, xd)
      + xdv
  )

  return xd


def integrate_xdd(
    sys: System,
    x: Transform,
    xd: Motion,
    xdd: Motion,
) -> Tuple[Transform, Motion]:
  """Updates position and velocity for a system time step given acceleration.

  Args:
    sys: System to forward propagate
    x: position
    xd: velocity
    xdd: acceleration

  Returns:
    x: updated position
    xd: updated velocity
  """

  xd = xd + xdd * sys.opt.timestep
  damp = Motion(vel=sys.vel_damping, ang=sys.ang_damping)
  xd = jax.tree.map(lambda d, x: jp.exp(d * sys.opt.timestep) * x, damp, xd)

  @jax.vmap
  def op(x, xd):
    pos = x.pos + xd.vel * sys.opt.timestep
    rot_at_ang_quat = math.ang_to_quat(xd.ang) * 0.5 * sys.opt.timestep
    rot, _ = math.normalize(x.rot + math.quat_mul(rot_at_ang_quat, x.rot))
    return Transform(pos=pos, rot=rot)

  x = op(x, xd)

  return x, xd


def project_xd(sys: System, x: Transform, x_prev: Transform) -> Motion:
  """Performs the position based dynamics velocity projection step.

  The velocity and angular velocity must respect the spatial and quaternion
  distance (respectively) between x and x_prev.

  Args:
    sys: The system definition
    x: The current transform
    x_prev: The transform at the previous step

  Returns:
    New state with velocity pinned to respect distance traveled since x_prev
  """

  @jax.vmap
  def op(x, x_prev):
    vel = (x.pos - x_prev.pos) / sys.opt.timestep
    dq = math.relative_quat(x_prev.rot, x.rot)
    ang = 2.0 * dq[1:] / sys.opt.timestep
    scale = jp.where(dq[0] >= 0.0, 1.0, -1.0)
    ang = scale * ang
    return Motion(vel=vel, ang=ang)

  return op(x, x_prev)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Tests for mjcf."""

from absl.testing import absltest
from brax import test_utils
from brax.io import mjcf
import mujoco
import numpy as np

assert_almost_equal = np.testing.assert_array_almost_equal


class MjcfTest(absltest.TestCase):

  def test_load_pendulum(self):
    sys = test_utils.load_fixture('triple_pendulum.xml')

    assert_almost_equal(sys.gravity, np.array([0, 0, -9.81]))

    # check links
    self.assertSequenceEqual(sys.link_names, ['body1', 'body2', 'body3'])
    assert_almost_equal(
        sys.link.transform.pos, np.array([[0, 0, 0], [0, 0.5, 0], [0, 0.5, 0]])
    )
    assert_almost_equal(
        sys.link.transform.rot,
        np.array([[1, 0, 0, 0], [1, 0, 0, 0], [1, 0, 0, 0]]),
    )
    assert_almost_equal(
        sys.link.inertia.i, np.tile(np.eye(3), (3, 1, 1)) * 0.009
    )
    assert_almost_equal(
        sys.link.inertia.transform.pos,
        np.array([[0, 0.5, 0], [0, 0.5, 0], [0, 0.5, 0]]),
    )
    assert_almost_equal(sys.link.inertia.mass, np.array([1, 1, 1]))

    # check static fields: link types, parents
    self.assertEqual(sys.link_types, '111')
    self.assertEqual(sys.link_parents, (-1, 0, 1))

  def test_load_ant(self):
    sys = test_utils.load_fixture('ant.xml')

    # check links
    self.assertSequenceEqual(
        sys.link_names,
        ['torso', 'aux_1', '', 'aux_2', '', 'aux_3', '', 'aux_4', ''],
    )

    # check static fields: link types, parents
    self.assertEqual(sys.link_types, 'f11111111')
    self.assertEqual(sys.link_parents, (-1, 0, 1, 0, 3, 0, 5, 0, 7))

  def test_load_humanoid(self):
    sys = test_utils.load_fixture('humanoid.xml')

    # check links
    self.assertSequenceEqual(
        sys.link_names,
        [
            'torso',
            'lwaist',
            'pelvis',
            'right_thigh',
            'right_shin',
            'left_thigh',
            'left_shin',
            'right_upper_arm',
            'right_lower_arm',
            'left_upper_arm',
            'left_lower_arm',
        ],
    )
    self.assertEqual(sys.link_types, 'f2131312121')

  def test_custom(self):
    sys = test_utils.load_fixture('capsule.xml')
    assert_almost_equal(sys.elasticity, [0.2, 0.1])

  def test_joint_ref_check(self):
    with self.assertRaisesRegex(NotImplementedError, '`ref` attribute'):
      sys = test_utils.load_fixture('nonzero_joint_ref.xml')
      mjcf.validate_model(sys.mj_model)

  def test_world_body_transform(self):
    sys = test_utils.load_fixture('world_body_transform.xml')
    # world body is in the right position/orientation
    r = 0.70710677
    assert_almost_equal(sys.geom_pos[1], np.array([1.0, 0.0, 0.0]))
    assert_almost_equal(sys.geom_quat[1], np.array([r, 0.0, r, 0.0]), 5)
    # child body is transformed wrt world body
    assert_almost_equal(
        sys.init_q, np.array([1.245, 0.0, 0.0, 0.5, 0.5, 0.5, -0.5])
    )

  def test_load_fluid_box(self):
    sys = test_utils.load_fixture('fluid_box.xml')
    assert_almost_equal(sys.density, 1.2)
    assert_almost_equal(sys.viscosity, 0.15)

  def test_load_fluid_ellipsoid(self):
    with self.assertRaisesRegex(
        NotImplementedError, 'Ellipsoid fluid model not implemented'
    ):
      sys = test_utils.load_fixture('fluid_ellipsoid.xml')
      mjcf.validate_model(sys.mj_model)

  def test_load_wind(self):
    with self.assertRaisesRegex(
        NotImplementedError, 'option.wind is not implemented'
    ):
      sys = test_utils.load_fixture('fluid_wind.xml')
      mjcf.validate_model(sys.mj_model)

  def test_world_fromto(self):
    """Tests that a world element with fromto does not break mjcf.load."""
    sys = test_utils.load_fixture('world_fromto.xml')
    mjcf.validate_model(sys.mj_model)

  def test_loads_different_transmission(self):
    """Tests that the brax model loads with different transmission types."""
    mj = test_utils.load_fixture_mujoco('ant.xml')
    mj.actuator_trntype[0] = mujoco.mjtTrn.mjTRN_SITE
    mjcf.load_model(mj)  # loads without raising an error

    with self.assertRaisesRegex(NotImplementedError, 'transmission types'):
      mjcf.validate_model(mj)  # raises an error

if __name__ == '__main__':
  absltest.main()



# Copyright 2024 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Standalone scoring script for Barkour.

Reads a csv containing timestamps and (x, y) positions of the robot's base, and
outputs the score the robot received on the Barkour obstacle course.

python score_barkour.py --log_file=data/barkour_run_0.csv
> 0.877
python score_barkour.py --log_file=data/barkour_run_0.csv --touched_jump
> 0.753
python score_barkour.py --log_file=data/barkour_run_1.csv
> 0.854
python score_barkour.py --log_file=data/barkour_run_2.csv
> 0.020
"""

import collections
import dataclasses
import math
import pprint

from absl import app
from absl import flags
import numpy as np


log_file_flag = flags.DEFINE_string(
    "log_file",
    None,
    "Location of file containing timestamps and robot positions.",
)
touched_jumpboard_flag = flags.DEFINE_bool(
    "touched_jump",
    False,
    "Whether robot made contact with the broad jump, failing that obstacle.",
)
# The distance to the gate at which the gate counts as reached.
SUCCESS_DISTANCE = 0.02  # meters
# Time penalty applied if the robot doesn't reach the end table.
MAX_TIME = 50  # seconds
PENALTY_PER_OBSTACLE = 0.1
# Nominal size in meters. This is the approximate length of a trajectory
# completing the obstacle, including approach distance.
NOMINAL_SIZE = {
    "start table": 1,
    "weave poles": 6,
    "a frame": 6,
    "jump": 4,
    "end table": 1,
}
TARGET_VELOCITY = 1.69  # meters/second
# Expected completion time for each obstacle.
ALLOTED_TIMES = {k: v / TARGET_VELOCITY for k, v in NOMINAL_SIZE.items()}
# The position (x,y) of the end table that the robot has to reach and stand on.
END_TABLE_POSITION = np.array((4.5, -4.5))
END_TABLE_HALF_WIDTH = 0.5
# Amount of time robot must stay on end table to complete course.
END_TABLE_COMPLETION_TIME = 5


# Each obstacle is composed of gates that the agent has to go through.
@dataclasses.dataclass(frozen=True)
class Gate:
  # (x, y) coordinates in meters
  left_pole: np.ndarray
  right_pole: np.ndarray

OBSTACLES_DICT = collections.OrderedDict([
    ("start table", [Gate([1.4, 0], [1.4, -0.7])]),
    (
        "weave poles",
        [
            Gate([1.7, 0], [1.7, -0.7]),
            Gate([2.5, -0.5], [1.7, -0.7]),
            Gate([2.5, -0.5], [3.2, -0.7]),
            Gate([3.8, -0.5], [3.2, -0.7]),
            Gate([3.8, -0.5], [4.3, -0.7]),
            Gate([5, -0.7], [4.3, -0.7]),
        ],
    ),
    (
        "a frame",
        [
            Gate([4.35, -2.0], [3.65, -1.1]),
            Gate([2.92, -3.0], [2.22, -2.1]),
            Gate([1.5, -4.0], [0.8, -3.1]),
        ],
    ),
    ("jump", [Gate([1.7, -4], [1.7, -5]), Gate([3.7, -4], [3.7, -5])]),
    ("end table", [Gate([4.1, -4], [4.1, -5])]),
])


def score_barkour(timestamps, positions, touched_jump):
  """Computes overall score and metrics for Barkour obstacle course.

  Args:
    timestamps: Sequence of timestamps (seconds) corresponding to positions. Can
      start at any value, as only differences are considered.
    positions: Array of (x, y) positions (meters) of the robot's base, relative
      to the Barkour scene. These will be checked against the positions of the
      Gates that make up the obstacles.
    touched_jump: Boolean indicating whether the robot touched the broad jump
      obstacle. If so, that obstacle counts as failed, even if the robot
      successfully completes the Gates.

  Returns:
    score: Barkour score in the range [0, 1].
    metrics: Dictionary containing extra information on the run.
  """
  completed_gates = {obs: 0 for obs in OBSTACLES_DICT}
  time_reached_end_table = 0
  time_at_end_table = 0
  trajectory_length = 0

  start_time = timestamps[0]
  last_robot_position = positions[0]

  for (timestamp, base_position) in zip(timestamps, positions):
    distance_moved = np.linalg.norm(last_robot_position - base_position)
    trajectory_length += distance_moved
    # Check if any of the obstacle gates are passed through.
    for key, gates_list in OBSTACLES_DICT.items():
      # Skip if the obstacle is already completed.
      if completed_gates[key] >= len(gates_list):
        continue
      active_gate = gates_list[completed_gates[key]]
      if _check_pass_through_gate(
          active_gate, base_position, last_robot_position):
        completed_gates[key] += 1
        # If it left the start table update the starting time.
        if key == "start table":
          start_time = timestamp

    if _robot_at_end_table(base_position):
      if time_reached_end_table == 0:
        time_reached_end_table = timestamp - 1e-6
      time_at_end_table = timestamp - time_reached_end_table
      if time_at_end_table >= END_TABLE_COMPLETION_TIME:
        break
    elif time_at_end_table > 0 and _robot_fell_outside_end_table(base_position):
      break

    last_robot_position = base_position

  time_spent = timestamp - start_time
  (allotted_course_time, num_obstacles_completed, progress_per_obstacle) = (
      _calculate_allotted_course_time(
          completed_gates, time_at_end_table, touched_jump
      )
  )
  agility_score = _calculate_agility_score(
      allotted_course_time,
      time_spent,
      num_obstacles_completed,
      time_at_end_table,
  )

  num_gates_completed = sum(completed_gates.values())
  total_gates = sum([len(x) for x in OBSTACLES_DICT.values()])
  course_time = time_spent - time_at_end_table
  metrics = {
      "gates_completed": num_gates_completed,
      "total_gates": total_gates,
      "progress_per_obstacle": progress_per_obstacle,
      "obstacles_completed": num_obstacles_completed,
      "time_at_end_table": time_at_end_table,
      "allotted_course_time_seconds": allotted_course_time,
      "final_run_time_seconds": course_time,
      "excess_time_seconds": course_time - allotted_course_time,
      "trajectory_length_meter": trajectory_length,
      "agility_score": agility_score
  }
  return agility_score, metrics


def _calculate_agility_score(
    allotted_course_time, time_spent, num_obstacles_completed, time_at_end_table
):
  penalty = PENALTY_PER_OBSTACLE * (
      len(OBSTACLES_DICT) - num_obstacles_completed)
  if time_at_end_table >= END_TABLE_COMPLETION_TIME:
    time_spent -= time_at_end_table
  else:
    time_spent = MAX_TIME + allotted_course_time
  return 1.0 - max(time_spent - allotted_course_time, 0) * 0.01 - penalty


def _calculate_allotted_course_time(
    completed_gates, time_at_end_table, touched_jump
):
  allotted_course_time = 0
  progress_per_obstacle = {}
  num_obstacles_completed = 0
  for obstacle, gates in OBSTACLES_DICT.items():
    progress_per_obstacle[obstacle] = completed_gates[
        obstacle] / len(gates)
    if obstacle == "jump" and touched_jump:
      progress_per_obstacle[obstacle] = 0
    if (obstacle == "end table" and time_at_end_table >= 5) or (
        obstacle != "end table" and
        progress_per_obstacle[obstacle] >= 1.0):
      num_obstacles_completed += 1
      allotted_course_time += ALLOTED_TIMES[obstacle]
  return allotted_course_time, num_obstacles_completed, progress_per_obstacle


def _robot_at_end_table(base_position):
  return max(np.abs(END_TABLE_POSITION - base_position)) < END_TABLE_HALF_WIDTH


def _robot_fell_outside_end_table(base_position):
  """Check if the robot is 110% outside the final table."""
  return (
      max(np.abs(END_TABLE_POSITION - base_position))
      > END_TABLE_HALF_WIDTH * 1.1
  )


def _check_pass_through_gate(gate, current_robot_position, last_robot_position):
  """Checks if the robot has passed through the gate or close enough to it."""
  distance_covered_since_last_step = np.linalg.norm(
      current_robot_position - last_robot_position
  )
  # Only do this check if the position jumped far enough to skip over the gate.
  if distance_covered_since_last_step >= 2 * SUCCESS_DISTANCE:
    return _check_2d_line_intersection(
        last_robot_position,
        current_robot_position,
        gate.left_pole,
        gate.right_pole,
    )
  # Otherwise we check the current distance to the gate line segment.
  gate_distance = _calculate_distance_to_line_segment(
      *current_robot_position, *gate.left_pole, *gate.right_pole)
  return gate_distance < SUCCESS_DISTANCE


def _calculate_distance_to_line_segment(x, y, x1, y1, x2, y2):
  """Calculates the shortest distance between a point and a line segment in 2D.

  Args:
    x: Point's x coordinate.
    y: Point's y coordinate.
    x1: Start of the line's x coordinate.
    y1: Start of the line's y coordinate.
    x2: End of the line's x coordinate.
    y2: End of the line's y coordinate.

  Returns:
    Shortest distance between the point and the line segment.
  """
  dist_x = x - x1
  dist_y = y - y1
  segment_x = x2 - x1
  segment_y = y2 - y1

  dot = dist_x * segment_x + dist_y * segment_y
  len_sq = segment_x * segment_x + segment_y * segment_y
  param = -1
  # Find the closest point in the segment.
  if len_sq > 0:
    param = dot / len_sq
  if param < 0:
    # The start of the segment is the closest point.
    point_x = x1
    point_y = y1
  elif param > 1:
    # The end of the segment is the closest point.
    point_x = x2
    point_y = y2
  else:
    # Linear interpolation to the closest point.
    point_x = x1 + param * segment_x
    point_y = y1 + param * segment_y

  return math.sqrt((x - point_x) * (x - point_x) + (y - point_y) *
                   (y - point_y))


def _orientation(p, q, r):
  """Find the orientation of an ordered triplet (p,q,r)."""
  orientation = (float(q[1] - p[1]) * (r[0] - q[0])) - (
      float(q[0] - p[0]) * (r[1] - q[1]))
  return np.sign(orientation)


def _check_2d_line_intersection(p1, q1, p2, q2):
  """Check if 2 lines intersect in 2D while ignoring colinear cases."""
  # Find the 4 orientations required.
  ori_1 = _orientation(p1, q1, p2)
  ori_2 = _orientation(p1, q1, q2)
  ori_3 = _orientation(p2, q2, p1)
  ori_4 = _orientation(p2, q2, q1)

  # General case.
  if ((ori_1 != ori_2) and (ori_3 != ori_4)):
    return True
  # We ignore colinear cases for the purpose of the barkour, because we check
  # the current distance of the robot to the gate.
  return False


def main(_):
  data = np.loadtxt(log_file_flag.value, delimiter=",")
  timestamps = data[:, 0]
  positions = data[:, 1:]
  score, metrics = score_barkour(
      timestamps, positions, touched_jumpboard_flag.value
  )
  print("Detailed metrics:")
  pprint.pprint(metrics)
  print(f"\nBarkour score: {score}\n")


if __name__ == "__main__":
  app.run(main)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Tests for actuators."""

from absl.testing import absltest
from absl.testing import parameterized
from brax import actuator
from brax import test_utils
from brax.generalized import pipeline as g_pipeline
from brax.positional import pipeline as p_pipeline
from brax.spring import pipeline as s_pipeline
import jax
from jax import numpy as jp
import mujoco
import numpy as np

assert_almost_equal = np.testing.assert_almost_equal


def _actuator_step(pipeline, sys, q, qd, act, dt, n):
  sys = sys.tree_replace({'opt.timestep': dt})
  def f(state, _):
    return jax.jit(pipeline.step)(sys, state, act), None

  state = jax.lax.scan(f, pipeline.init(sys, q, qd), (), n)[0]
  return state.q, state.qd


class ActuatorTest(parameterized.TestCase):

  @parameterized.parameters(
      (g_pipeline, 0.01, 100, 5),
      (s_pipeline, 0.001, 1000, 3),
      (p_pipeline, 0.001, 1000, 3),
  )
  def test_motor(self, pipeline, dt, n, decimal):
    """Tests a single pendulum with motor actuator."""
    sys = test_utils.load_fixture('single_pendulum_motor.xml')
    mj_model = test_utils.load_fixture_mujoco('single_pendulum_motor.xml')
    mj_data = mujoco.MjData(mj_model)
    q, qd = jp.array(mj_data.qpos), jp.array(mj_data.qvel)

    tau = jp.array([0.5 * 9.81])  # -mgl sin(theta)
    act = jp.array([1.0 / 150.0 * 0.5 * 9.81])
    tau2 = actuator.to_tau(sys, act, q, qd)
    np.testing.assert_array_almost_equal(tau, tau2, 5)

    q2, qd2 = _actuator_step(pipeline, sys, q, qd, act=act, dt=dt, n=n)
    np.testing.assert_array_almost_equal(q2, jp.array([0]), decimal=decimal)
    np.testing.assert_array_almost_equal(qd2, jp.array([0]), decimal=decimal)

  def test_position(self):
    """Tests a single pendulum with position actuator."""
    sys = test_utils.load_fixture('single_pendulum_position.xml')
    mj_model = test_utils.load_fixture_mujoco('single_pendulum_position.xml')
    mj_data = mujoco.MjData(mj_model)
    q, qd = jp.array(mj_data.qpos), jp.array(mj_data.qvel)
    theta = jp.pi / 2.0  # pendulum is vertical
    q = jp.array([theta])

    # a position actuator at the bottom should not move the pendulum
    act = jp.array([theta])
    tau = actuator.to_tau(sys, act, q, qd)
    np.testing.assert_array_almost_equal(tau, jp.array([0]), 5)

    # put the pendulum into the horizontal position with the positional actuator
    # we know alpha = -2 theta/t^2, and I * alpha = (act-bias) * gear, and we
    # then solve for act. Semi-implicit Euler leaves off a factor of two.
    act = jp.array([-(theta * 0.5**2) / (0.01**2 * 10.0) + theta])
    q2, _ = _actuator_step(g_pipeline, sys, q, qd, act=act, dt=0.01, n=1)
    np.testing.assert_array_almost_equal(q2, jp.array([0]), 1)

  def test_velocity(self):
    """Tests a single pendulum with velocity actuator."""
    sys = test_utils.load_fixture('single_pendulum_velocity.xml')
    mj_model = test_utils.load_fixture_mujoco('single_pendulum_velocity.xml')
    mj_data = mujoco.MjData(mj_model)
    q, qd = jp.array(mj_data.qpos), jp.array(mj_data.qvel)
    theta = jp.pi / 2.0  # pendulum is vertical
    q = jp.array([theta])

    act = jp.array([0])
    tau = actuator.to_tau(sys, act, q, qd)
    np.testing.assert_array_almost_equal(tau, jp.array([0]), 5)

    # set the act to rotate at 1/s
    act = jp.array([1])
    _, qd = _actuator_step(g_pipeline, sys, q, qd, act=act, dt=0.001, n=200)
    np.testing.assert_array_almost_equal(qd, jp.array([1]), 3)

  def test_force_limitted(self):
    """Tests that forcerange limits work on actuators."""
    sys = test_utils.load_fixture('single_pendulum_position_frclimit.xml')
    mj_model = test_utils.load_fixture_mujoco(
        'single_pendulum_position_frclimit.xml'
    )
    mj_data = mujoco.MjData(mj_model)
    q, qd = jp.array(mj_data.qpos), jp.array(mj_data.qvel)

    for act, frclimit in [(1000, 3.1), (-1000, -2.5)]:
      act = jp.array([act])
      tau = actuator.to_tau(sys, act, q, qd)
      # test that tau matches frclimit * 10, since gear=10
      self.assertEqual(tau[0], frclimit * 10)
      # test that tau matches MJ qfrc_actuator
      mj_data.ctrl = act
      mujoco.mj_step(mj_model, mj_data)
      self.assertEqual(tau[0], mj_data.qfrc_actuator)

  @parameterized.parameters((g_pipeline,), (s_pipeline,), (p_pipeline,))
  def test_three_link_pendulum(self, pipeline):
    """Tests a three link pendulum with a motor actuator."""
    sys = test_utils.load_fixture('triple_pendulum_motor.xml')
    mj_model = test_utils.load_fixture_mujoco('triple_pendulum_motor.xml')
    mj_data = mujoco.MjData(mj_model)
    q, qd = jp.array(mj_data.qpos), jp.array(mj_data.qvel)
    theta = jp.pi / 2.0  # pendulum is vertical
    q = jp.array([theta, 0.0, 0.0])

    # test that with zero action the triple pendulum does not move
    act = jp.array([0] * 3)
    q1, qd1 = _actuator_step(pipeline, sys, q, qd, act=act, dt=0.01, n=1)
    np.testing.assert_array_almost_equal(q1, q, 2)
    np.testing.assert_array_almost_equal(qd1, jp.zeros_like(qd1), 2)

    # now test that each torque results in directionally accurate q and qd
    act = 1.0 / 150.0 * jp.array([0, -10, 10])
    q2, qd2 = _actuator_step(pipeline, sys, q, qd, act=act, dt=1e-3, n=1)
    self.assertAlmostEqual(q2[0], q[0], 2)
    self.assertLess(q2[1], q[1])
    self.assertGreater(q2[2], q[2])
    self.assertLess(qd2[1], -0.2)
    self.assertGreater(qd2[2], -0.2)

  @parameterized.parameters(
      ('single_spherical_pendulum_motor.xml',), 'single_pendulum_motor.xml'
  )
  def test_spherical_pendulum_mj_generalized(self, config):
    """Tests a spherical pendulum with a motor actuator against mujoco."""
    # Note that positional actuators between mujoco and brax are not equivalent
    # so we only test motor actuators here.
    sys = test_utils.load_fixture(config)
    mj_model = test_utils.load_fixture_mujoco(config)
    mj_data = mujoco.MjData(mj_model)
    q, qd = jp.asarray(mj_data.qpos), jp.asarray(mj_data.qvel)

    act = jp.array([0.1, 0.2, 0.3])[: sys.act_size()]
    mj_data.ctrl = act
    for _ in range(1000):
      mujoco.mj_step(mj_model, mj_data)
    mq, mqd = jp.asarray(mj_data.qpos), jp.asarray(mj_data.qvel)

    gq, gqd = _actuator_step(
        g_pipeline, sys, q, qd, act=act, dt=sys.opt.timestep, n=1000
    )
    np.testing.assert_array_almost_equal(gq, mq, 3)
    np.testing.assert_array_almost_equal(gqd, mqd, 3)

  @parameterized.parameters(
      'single_pendulum_position.xml',
      'single_pendulum_motor.xml',
      'single_spherical_pendulum_position.xml',
  )
  def test_single_pendulum_spring_positional(self, config):
    sys = test_utils.load_fixture(config)
    act = jp.array([0.05, 0.1, 0.15])[: sys.act_size()]

    q, qd = sys.init_q, jp.zeros(sys.qd_size())

    sq, sqd = _actuator_step(
        s_pipeline, sys, q, qd, act=act, dt=sys.opt.timestep, n=500
    )
    pq, pqd = _actuator_step(
        p_pipeline, sys, q, qd, act=act, dt=sys.opt.timestep, n=500
    )
    np.testing.assert_array_almost_equal(sq, pq, 2)
    np.testing.assert_array_almost_equal(sqd, pqd, 2)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Function to load MuJoCo mjcf format to Brax model."""

import itertools
from typing import Dict, Optional, Tuple, Union
from xml.etree import ElementTree

from brax import math
from brax.base import (
    Actuator,
    DoF,
    Inertia,
    Link,
    Motion,
    System,
    Transform,
)
from etils import epath
import jax
from jax import numpy as jp
import mujoco
from mujoco import mjx
import numpy as np


def _transform_do(
    parent_pos: np.ndarray, parent_quat: np.ndarray, pos: np.ndarray,
    quat: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
  pos = parent_pos + math.rotate_np(pos, parent_quat)
  rot = math.quat_mul_np(parent_quat, quat)
  return pos, rot


def _offset(
    elem: ElementTree.Element, parent_pos: np.ndarray, parent_quat: np.ndarray):
  """Offsets an element."""
  pos = elem.attrib.get('pos', '0 0 0')
  quat = elem.attrib.get('quat', '1 0 0 0')
  pos = np.fromstring(pos, sep=' ')
  quat = np.fromstring(quat, sep=' ')
  fromto = elem.attrib.get('fromto', None)
  if fromto:
    # fromto attributes are not compatible with pos/quat attributes
    from_pos = np.fromstring(' '.join(fromto.split(' ')[0:3]), sep=' ')
    to_pos = np.fromstring(' '.join(fromto.split(' ')[3:6]), sep=' ')
    from_pos, _ = _transform_do(parent_pos, parent_quat, from_pos, quat)
    to_pos, _ = _transform_do(parent_pos, parent_quat, to_pos, quat)
    fromto = ' '.join('%f' % i for i in np.concatenate([from_pos, to_pos]))
    elem.attrib['fromto'] = fromto
    return
  pos, quat = _transform_do(parent_pos, parent_quat, pos, quat)
  pos = ' '.join('%f' % i for i in pos)
  quat = ' '.join('%f' % i for i in quat)
  elem.attrib['pos'] = pos
  elem.attrib['quat'] = quat


def _fuse_bodies(elem: ElementTree.Element):
  """Fuses together parent child bodies that have no joint."""

  for child in list(elem):  # we will modify elem children, so make a copy
    _fuse_bodies(child)
    # this only applies to bodies with no joints
    if child.tag != 'body':
      continue
    if child.find('joint') is not None or child.find('freejoint') is not None:
      continue
    cpos = child.attrib.get('pos', '0 0 0')
    cpos = np.fromstring(cpos, sep=' ')
    cquat = child.attrib.get('quat', '1 0 0 0')
    cquat = np.fromstring(cquat, sep=' ')
    for grandchild in child:
      # TODO: might need to offset more than just these tags
      if (
          grandchild.tag in ('body', 'geom', 'site', 'camera')
          and (cpos != 0).any()
      ):
        _offset(grandchild, cpos, cquat)
      elem.append(grandchild)
    elem.remove(child)


def _get_meshdir(elem: ElementTree.Element) -> Union[str, None]:
  """Gets the mesh directory specified by the mujoco compiler tag."""
  elems = list(elem.iter('compiler'))
  return elems[0].get('meshdir') if elems else None


def _find_assets(
    elem: ElementTree.Element,
    path: epath.Path,
    meshdir: Optional[str],
) -> Dict[str, bytes]:
  """Loads assets from an xml given a base path."""
  assets = {}
  path = path if path.is_dir() else path.parent
  fname = elem.attrib.get('file') or elem.attrib.get('filename')
  if fname and fname.endswith('.xml'):
    # an asset can be another xml!  if so, we must traverse it, too
    asset = (path / fname).read_text()
    asset_xml = ElementTree.fromstring(asset)
    _fuse_bodies(asset_xml)
    asset_meshdir = _get_meshdir(asset_xml)
    assets[fname] = ElementTree.tostring(asset_xml)
    assets.update(_find_assets(asset_xml, path, asset_meshdir))
  elif fname:
    # mesh, png, etc
    path = path / meshdir if meshdir else path
    assets[fname] = (path / fname).read_bytes()

  for child in list(elem):
    assets.update(_find_assets(child, path, meshdir))

  return assets


def _get_name(mj: mujoco.MjModel, i: int) -> str:
  names = mj.names[i:].decode('utf-8')
  return names[: names.find('\x00')]


def _check_custom(mj: mujoco.MjModel, custom: Dict[str, np.ndarray]) -> None:
  """Validates fields in custom."""
  if not (
      0 <= custom['spring_mass_scale'] <= 1
      and 0 <= custom['spring_inertia_scale'] <= 1
  ):
    raise ValueError('Spring inertia and mass scale must be in [0, 1].')
  if 'init_qpos' in custom and custom['init_qpos'].shape[0] != mj.nq:
    size = custom['init_qpos'].shape[0]
    raise ValueError(
        f'init_qpos had length {size} but expected length {mj.nq}.'
    )


def _get_custom(mj: mujoco.MjModel) -> Dict[str, np.ndarray]:
  """Gets custom mjcf parameters for brax, with defaults."""
  default = {
      'ang_damping': (0.0, None),
      'vel_damping': (0.0, None),
      'baumgarte_erp': (0.1, None),
      'spring_mass_scale': (0.0, None),
      'spring_inertia_scale': (0.0, None),
      'joint_scale_pos': (0.5, None),
      'joint_scale_ang': (0.2, None),
      'collide_scale': (1.0, None),
      'matrix_inv_iterations': (10, None),
      'solver_maxls': (20, None),
      'elasticity': (0.0, 'geom'),
      'constraint_stiffness': (2000.0, 'body'),
      'constraint_limit_stiffness': (1000.0, 'body'),
      'constraint_ang_damping': (0.0, 'body'),
      'constraint_vel_damping': (0.0, 'body'),
  }

  # add user provided overrides to the defaults
  for i, ni in enumerate(mj.name_numericadr):
    nsize = mj.numeric_size[i]
    name = _get_name(mj, ni)
    val = mj.numeric_data[mj.numeric_adr[i] : mj.numeric_adr[i] + nsize]
    typ = default[name][1] if name in default else None
    default[name] = (val, typ)

  # gather custom overrides with correct sizes
  custom = {}
  for name, (val, typ) in default.items():
    val = np.array([val])
    size = {
        'body': mj.nbody - 1,  # ignore the world body
        'geom': mj.ngeom,
    }.get(typ, val.shape[-1])
    if val.shape[-1] != size and val.shape[-1] > 1:
      # the provided shape does not match against our default size
      raise ValueError(
          f'"{name}" custom arg needed {size} values for the "{typ}" type, '
          f'but got {val.shape[-1]} values.'
      )
    elif val.shape[-1] != size and val.shape[-1] == 1:
      val = np.repeat(val, size)
    val = val.squeeze() if not typ else val.reshape(size)
    if typ == 'body':
      # pad one value for the world body, which gets dropped at Link creation
      val = np.concatenate([[val[0]], val])
    custom[name] = val

  # get tuple custom overrides
  for i, ni in enumerate(mj.name_tupleadr):
    start, end = mj.tuple_adr[i], mj.tuple_adr[i] + mj.tuple_size[i]
    objtype = mj.tuple_objtype[start:end]
    name = _get_name(mj, ni)
    if not all(objtype[0] == objtype):
      raise NotImplementedError(
          f'All tuple elements "{name}" should have the same object type.'
      )
    if objtype[0] not in [1, 5]:
      raise NotImplementedError(
          f'Custom tuple "{name}" with objtype=={objtype[0]} is not supported.'
      )
    typ = {1: 'body', 5: 'geom'}[objtype[0]]
    if name in default and default[name][1] != typ:
      raise ValueError(
          f'Custom tuple "{name}" is expected to be associated with'
          f' the {default[name][1]} objtype.'
      )

    size = {1: mj.nbody, 5: mj.ngeom}[objtype[0]]
    default_val, _ = default.get(name, (0.0, None))
    arr = np.repeat(default_val, size)
    objid = mj.tuple_objid[start:end]
    objprm = mj.tuple_objprm[start:end]
    arr[objid] = objprm
    custom[name] = arr

  _check_custom(mj, custom)
  return custom


def validate_model(mj: mujoco.MjModel) -> None:
  """Checks if a MuJoCo model is compatible with brax physics pipelines."""
  if mj.opt.integrator != 0:
    raise NotImplementedError('Only euler integration is supported.')
  if mj.opt.cone != 0:
    raise NotImplementedError('Only pyramidal cone friction is supported.')
  if (mj.geom_fluid != 0).any():
    raise NotImplementedError('Ellipsoid fluid model not implemented.')
  if mj.opt.wind.any():
    raise NotImplementedError('option.wind is not implemented.')
  if mj.opt.impratio != 1:
    raise NotImplementedError('Only impratio=1 is supported.')

  # actuators
  if any(i not in [0, 1] for i in mj.actuator_biastype):
    raise NotImplementedError('Only actuator_biastype in [0, 1] are supported.')
  if any(i != 0 for i in mj.actuator_gaintype):
    raise NotImplementedError('Only actuator_gaintype in [0] is supported.')
  if not (mj.actuator_trntype == 0).all():
    raise NotImplementedError(
        'Only joint transmission types are supported for actuators.'
    )

  # solver parameters
  if (mj.geom_solmix[0] != mj.geom_solmix).any():
    raise NotImplementedError('geom_solmix parameter not supported.')
  if (mj.geom_priority[0] != mj.geom_priority).any():
    raise NotImplementedError('geom_priority parameter not supported.')

  # check joints
  q_width = {0: 7, 1: 4, 2: 1, 3: 1}
  non_free = np.concatenate([[j != 0] * q_width[j] for j in mj.jnt_type])
  if mj.qpos0[non_free].any():
    raise NotImplementedError(
        'The `ref` attribute on joint types is not supported.')

  for _, group in itertools.groupby(
      zip(mj.jnt_bodyid, mj.jnt_pos), key=lambda x: x[0]
  ):
    position = np.array([p for _, p in group])
    if not (position == position[0]).all():
      raise RuntimeError('invalid joint stack: only one joint position allowed')

  # check dofs
  jnt_range = mj.jnt_range.copy()
  jnt_range[~(mj.jnt_limited == 1), :] = np.array([-np.inf, np.inf])
  for typ, limit, stiffness in zip(
      mj.jnt_type, jnt_range, mj.jnt_stiffness
  ):
    if typ == 0:
      if stiffness > 0:
        raise RuntimeError('brax does not support stiffness for free joints')
    elif typ == 1:
      if np.any(~np.isinf(limit)):
        raise RuntimeError('brax does not support joint ranges for ball joints')
    elif typ in (2, 3):
      continue
    else:
      raise RuntimeError(f'invalid joint type: {typ}')

  for _, group in itertools.groupby(
      zip(mj.jnt_bodyid, mj.jnt_type), key=lambda x: x[0]
  ):
    typs = [t for _, t in group]
    if len(typs) == 1 and typs[0] == 0:
      continue  # free
    elif 0 in typs:
      raise RuntimeError('invalid joint stack: cannot stack free joints')
    elif 1 in typs:
      raise NotImplementedError('ball joints not supported')

  # check collision geometries
  for i, typ in enumerate(mj.geom_type):
    mask = mj.geom_contype[i] | mj.geom_conaffinity[i] << 32
    if typ == 5:  # Cylinder
      _, halflength = mj.geom_size[i, 0:2]
      if halflength > 0.001 and mask > 0:
        raise NotImplementedError(
            'Cylinders of half-length>0.001 are not supported for collision.'
        )


def load_model(mj: mujoco.MjModel) -> System:
  """Creates a brax system from a MuJoCo model."""
  custom = _get_custom(mj)

  # create links
  joint_positions = [np.array([0.0, 0.0, 0.0])]
  for _, group in itertools.groupby(
      zip(mj.jnt_bodyid, mj.jnt_pos), key=lambda x: x[0]
  ):
    position = np.array([p for _, p in group])
    joint_positions.append(position[0])
  joint_position = np.array(joint_positions)
  identity = np.tile(np.array([1.0, 0.0, 0.0, 0.0]), (mj.nbody, 1))
  link = Link(  # pytype: disable=wrong-arg-types  # jax-ndarray
      transform=Transform(pos=mj.body_pos, rot=mj.body_quat),  # pytype: disable=wrong-arg-types  # jax-ndarray
      inertia=Inertia(  # pytype: disable=wrong-arg-types  # jax-ndarray
          transform=Transform(pos=mj.body_ipos, rot=mj.body_iquat),  # pytype: disable=wrong-arg-types  # jax-ndarray
          i=np.array([np.diag(i) for i in mj.body_inertia]),
          mass=mj.body_mass,
      ),
      invweight=mj.body_invweight0[:, 0],
      joint=Transform(pos=joint_position, rot=identity),  # pytype: disable=wrong-arg-types  # jax-ndarray
      constraint_stiffness=custom['constraint_stiffness'],
      constraint_vel_damping=custom['constraint_vel_damping'],
      constraint_limit_stiffness=custom['constraint_limit_stiffness'],
      constraint_ang_damping=custom['constraint_ang_damping'],
  )
  # skip link 0 which is the world body in mujoco
  # copy to avoid writing to mj model
  link = jax.tree.map(lambda x: x[1:].copy(), link)

  # create dofs
  mj.jnt_range[~(mj.jnt_limited == 1), :] = np.array([-np.inf, np.inf])
  motions, limits, stiffnesses = [], [], []
  for typ, axis, limit, stiffness in zip(
      mj.jnt_type, mj.jnt_axis, mj.jnt_range, mj.jnt_stiffness
  ):
    if typ == 0:
      motion = Motion(ang=np.eye(6, 3, -3), vel=np.eye(6, 3))
      limit = np.array([-np.inf] * 6), np.array([np.inf] * 6)
      stiffness = np.zeros(6)
    elif typ == 1:
      motion = Motion(ang=np.eye(3), vel=np.zeros((3, 3)))
      limit = np.array([-np.inf] * 3), np.array([np.inf] * 3)
      stiffness = np.zeros(3)
    elif typ == 2:
      motion = Motion(ang=np.zeros((1, 3)), vel=axis.reshape((1, 3)))
      limit = limit[0:1], limit[1:2]
      stiffness = np.array([stiffness])
    elif typ == 3:
      motion = Motion(ang=axis.reshape((1, 3)), vel=np.zeros((1, 3)))
      limit = limit[0:1], limit[1:2]
      stiffness = np.array([stiffness])
    else:
      # invalid joint type
      continue
    motions.append(motion)
    limits.append(limit)
    stiffnesses.append(stiffness)
  motion = jax.tree.map(lambda *x: np.concatenate(x), *motions)

  limit = None
  if np.any(mj.jnt_limited):
    limit = jax.tree.map(lambda *x: np.concatenate(x), *limits)
  stiffness = np.concatenate(stiffnesses)
  solver_params_jnt = np.concatenate((mj.jnt_solref, mj.jnt_solimp), axis=1)
  solver_params_dof = solver_params_jnt[mj.dof_jntid]

  dof = DoF(  # pytype: disable=wrong-arg-types
      motion=motion,
      armature=mj.dof_armature,
      stiffness=stiffness,
      damping=mj.dof_damping,
      limit=limit,
      invweight=mj.dof_invweight0,
      solver_params=solver_params_dof,
  )

  # create actuators
  # TODO: swap brax actuation for mjx actuation model.
  ctrl_range = mj.actuator_ctrlrange
  ctrl_range[~(mj.actuator_ctrllimited == 1), :] = np.array([-np.inf, np.inf])
  force_range = mj.actuator_forcerange
  force_range[~(mj.actuator_forcelimited == 1), :] = np.array([-np.inf, np.inf])
  bias_q = mj.actuator_biasprm[:, 1] * (mj.actuator_biastype != 0)
  bias_qd = mj.actuator_biasprm[:, 2] * (mj.actuator_biastype != 0)
  # mask actuators since brax only supports joint transmission types
  act_mask = mj.actuator_trntype == mujoco.mjtTrn.mjTRN_JOINT
  trnid = mj.actuator_trnid[act_mask, 0].astype(np.uint32)
  q_id = mj.jnt_qposadr[trnid]
  qd_id = mj.jnt_dofadr[trnid]
  act_kwargs = {
      'gain': mj.actuator_gainprm[:, 0],
      'gear': mj.actuator_gear[:, 0],
      'ctrl_range': ctrl_range,
      'force_range': force_range,
      'bias_q': bias_q,
      'bias_qd': bias_qd,
  }
  act_kwargs = jax.tree.map(lambda x: x[act_mask], act_kwargs)

  actuator = Actuator(  # pytype: disable=wrong-arg-types
      q_id=q_id,
      qd_id=qd_id,
      **act_kwargs
  )

  # create non-pytree params.  these do not live on device directly, and they
  # cannot be differentiated, but they do change the emitted control flow
  link_names = [_get_name(mj, i) for i in mj.name_bodyadr[1:]]
  # convert stacked joints to 1, 2, or 3
  link_types = ''
  for _, group in itertools.groupby(
      zip(mj.jnt_bodyid, mj.jnt_type), key=lambda x: x[0]
  ):
    typs = [t for _, t in group]
    if len(typs) == 1 and typs[0] == 0:  # free
      typ = 'f'
    elif 0 in typs or 1 in typs:
      # invalid joint stack
      continue
    else:
      typ = str(len(typs))
    link_types += typ
  link_parents = tuple(mj.body_parentid - 1)[1:]

  # mujoco stores free q in world frame, so clear link transform for free links
  # TODO: make this work for non-fused mj models
  if 'f' in link_types:
    free_idx = np.array([i for i, typ in enumerate(link_types) if typ == 'f'])
    link.transform.pos[free_idx] = np.zeros(3)
    link.transform.rot[free_idx] = np.array([1.0, 0.0, 0.0, 0.0])

  mjx_model = mjx.put_model(mj)

  sys = System(  # pytype: disable=wrong-arg-types  # jax-ndarray
      gravity=mj.opt.gravity,
      viscosity=mj.opt.viscosity,
      density=mj.opt.density,
      elasticity=custom['elasticity'],
      link=link,
      dof=dof,
      actuator=actuator,
      init_q=custom['init_qpos'] if 'init_qpos' in custom else mj.qpos0,
      vel_damping=custom['vel_damping'],
      ang_damping=custom['ang_damping'],
      baumgarte_erp=custom['baumgarte_erp'],
      spring_mass_scale=custom['spring_mass_scale'],
      spring_inertia_scale=custom['spring_inertia_scale'],
      joint_scale_ang=custom['joint_scale_ang'],
      joint_scale_pos=custom['joint_scale_pos'],
      collide_scale=custom['collide_scale'],
      enable_fluid=(mj.opt.viscosity > 0) | (mj.opt.density > 0),
      link_names=link_names,
      link_types=link_types,
      link_parents=link_parents,
      matrix_inv_iterations=int(custom['matrix_inv_iterations']),
      solver_iterations=mj.opt.iterations,
      solver_maxls=int(custom['solver_maxls']),
      mj_model=mj,
      **mjx_model.__dict__,
  )

  sys = jax.tree.map(jp.array, sys)

  return sys


def fuse_bodies(xml: str):
  """Fuses together parent child bodies that have no joint."""
  xml = ElementTree.fromstring(xml)
  _fuse_bodies(xml)
  return ElementTree.tostring(xml, encoding='unicode')


def loads(xml: str, asset_path: Union[str, epath.Path, None] = None) -> System:
  """Loads a brax system from a MuJoCo mjcf xml string."""
  elem = ElementTree.fromstring(xml)
  _fuse_bodies(elem)
  assets = {}
  if asset_path is not None:
    meshdir = _get_meshdir(elem)
    asset_path = epath.Path(asset_path)
    assets = _find_assets(elem, asset_path, meshdir)
  xml = ElementTree.tostring(elem, encoding='unicode')
  mj = mujoco.MjModel.from_xml_string(xml, assets=assets)

  return load_model(mj)


def load_mjmodel(path: Union[str, epath.Path]) -> mujoco.MjModel:
  """Loads an mj model from a MuJoCo mjcf file path."""
  elem = ElementTree.fromstring(epath.Path(path).read_text())
  _fuse_bodies(elem)
  meshdir = _get_meshdir(elem)
  assets = _find_assets(elem, epath.Path(path), meshdir)
  xml = ElementTree.tostring(elem, encoding='unicode')
  mj = mujoco.MjModel.from_xml_string(xml, assets=assets)
  return mj


def load(path: Union[str, epath.Path]):
  """Loads a brax system from a MuJoCo mjcf file path."""
  mj = load_mjmodel(path)
  return load_model(mj)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import, g-importing-member
"""Functions for constraint satisfaction."""
from typing import Tuple

from brax import contact
from brax import math
from brax import scan
from brax.base import Motion, System, Transform
from brax.generalized.base import State
import jax
from jax import numpy as jp
import jaxopt


def _imp_aref(
    params: jax.Array, pos: jax.Array, vel: jax.Array
) -> Tuple[jax.Array, jax.Array]:
  """Calculates impedance and offset acceleration in constraint frame.

  Args:
    params: solver params
    pos: position in constraint frame
    vel: velocity in constraint frame

  Returns:
    imp: constraint impedance
    Aref: offset acceleration in constraint frame
  """
  # this formulation corresponds to the parameterization described here:
  # https://mujoco.readthedocs.io/en/latest/modeling.html#solver-parameters
  timeconst, dampratio, dmin, dmax, width, mid, power = params

  imp_x = jp.abs(pos) / width
  imp_a = (1.0 / jp.power(mid, power - 1)) * jp.power(imp_x, power)
  imp_b = 1 - (1.0 / jp.power(1 - mid, power - 1)) * jp.power(1 - imp_x, power)
  imp_y = jp.where(imp_x < mid, imp_a, imp_b)
  imp = dmin + imp_y * (dmax - dmin)
  imp = jp.clip(imp, dmin, dmax)
  imp = jp.where(imp_x > 1.0, dmax, imp)

  b = 2 / (dmax * timeconst)
  k = 1 / (dmax * dmax * timeconst * timeconst * dampratio * dampratio)

  # See https://mujoco.readthedocs.io/en/latest/modeling.html#solver-parameters
  stiffness, damping = params[:2]
  b = jp.where(damping <= 0, -damping / dmax, b)
  k = jp.where(stiffness <= 0, -stiffness / (dmax * dmax), k)

  aref = -b * vel - k * imp * pos

  return imp, aref


def point_jacobian(
    sys: System,
    com: jax.Array,
    cdof: Motion,
    pos: jax.Array,
    link_idx: jax.Array,
) -> Motion:
  """Calculates the jacobian of a point on a link.

  Args:
    sys: a brax system
    com: center of mass position
    cdof: dofs in com frame
    pos: position in world frame to calculate the jacobian
    link_idx: index of link frame to transform point jacobian

  Returns:
    pt: point jacobian
  """
  # backward scan up tree: build the link mask corresponding to link_idx
  def mask_fn(mask_child, link):
    mask = link == link_idx
    if mask_child is not None:
      mask += mask_child
    return mask

  mask = scan.tree(sys, mask_fn, 'l', jp.arange(sys.num_links()), reverse=True)
  cdof = jax.vmap(lambda a, b: a * b)(cdof, jp.take(mask, sys.dof_link()))
  off = Transform.create(pos=pos - com[link_idx])
  return off.vmap(in_axes=(None, 0)).do(cdof)


def jac_limit(
    sys: System, state: State
) -> Tuple[jax.Array, jax.Array, jax.Array]:
  """Calculates the jacobian for angle limits in dof frame.

  Args:
    sys: a brax system
    state: generalized state

  Returns:
    jac: the angle limit jacobian
    pos: angle in constraint frame
    diag: approximate diagonal of A matrix
  """
  if sys.dof.limit is None:
    return jp.zeros((0, sys.qd_size())), jp.zeros((0,)), jp.zeros((0,))

  # determine q and qd indices for non-free joints
  q_idx, qd_idx = sys.q_idx('123'), sys.qd_idx('123')

  pos_min = state.q[q_idx] - sys.dof.limit[0][qd_idx]
  pos_max = sys.dof.limit[1][qd_idx] - state.q[q_idx]
  pos = jp.minimum(jp.minimum(pos_min, pos_max), 0)

  side = ((pos_min < pos_max) * 2 - 1) * (pos < 0)
  jac = jax.vmap(jp.multiply)(jp.eye(sys.qd_size())[qd_idx], side)
  params = sys.dof.solver_params[qd_idx]
  imp, aref = jax.vmap(_imp_aref)(params, pos, jac @ state.qd)
  diag = sys.dof.invweight[qd_idx] * (pos < 0) * (1 - imp) / (imp + 1e-8)
  aref = jax.vmap(lambda x, y: x * y)(aref, (pos < 0))

  return jac, diag, aref


def jac_contact(
    sys: System, state: State
) -> Tuple[jax.Array, jax.Array, jax.Array]:
  """Calculates the jacobian for contact constraints.

  Args:
    sys: the brax system
    state: generalized state

  Returns:
    jac: the contact jacobian
    pos: contact position in constraint frame
    diag: approximate diagonal of A matrix
  """
  c = contact.get(sys, state.x)

  if c is None:
    return jp.zeros((0, sys.qd_size())), jp.zeros((0,)), jp.zeros((0,))

  def row_fn(c):
    link_a, link_b = c.link_idx
    a = point_jacobian(sys, state.root_com, state.cdof, c.pos, link_a)
    b = point_jacobian(sys, state.root_com, state.cdof, c.pos, link_b)
    diff = b.vel - a.vel

    # 4 pyramidal friction directions
    jac = []
    for d in -c.frame[1:]:
      for f in [-c.friction[0], c.friction[0]]:
        jac.append(diff @ (d * f + c.frame[0]))

    jac = jp.stack(jac)
    pos = jp.tile(c.dist, 4)
    solver_params = jp.concatenate([c.solref, c.solimp])
    imp, aref = _imp_aref(solver_params, pos, jac @ state.qd)
    t = sys.link.invweight[link_a] * (link_a > -1) + sys.link.invweight[link_b]
    diag = jp.tile(t + c.friction[0] * c.friction[0] * t, 4)
    diag *= 2 * c.friction[0] * c.friction[0] * (1 - imp) / (imp + 1e-8)

    return jax.tree.map(lambda x: x * (c.dist < 0), (jac, diag, aref))

  return jax.tree.map(jp.concatenate, jax.vmap(row_fn)(c))


def jacobian(sys: System, state: State) -> State:
  """Calculates the constraint jacobian, position, and A matrix diagonal.

  Args:
    sys: a brax system
    state: generalized state

  Returns:
    state: generalized state with jac, pos, diag, aref updated
  """
  jpds = jac_contact(sys, state), jac_limit(sys, state)
  jac, diag, aref = jax.tree.map(lambda *x: jp.concatenate(x), *jpds)
  return state.replace(con_jac=jac, con_diag=diag, con_aref=aref)


def force(sys: System, state: State) -> jax.Array:
  """Calculates forces that satisfy joint, collision constraints.

  Args:
    sys: a brax system
    state: generalized state

  Returns:
    qf_constraint: (qd_size,) constraint force
  """
  if state.con_jac.shape[0] == 0:
    return jp.zeros(sys.qd_size())

  # calculate A matrix and b vector
  a = state.con_jac @ state.mass_mx_inv @ state.con_jac.T
  a = a + jp.diag(state.con_diag)
  b = state.con_jac @ state.mass_mx_inv @ state.qf_smooth - state.con_aref

  # solve for forces in constraint frame, Ax + b = 0 s.t. x >= 0
  def objective(x):
    residual = a @ x + b
    return jp.sum(0.5 * residual**2)

  # profiling a physics step, most of the time is spent running this solver:
  #
  # there might still be some opportunities to speed this up.  consider that
  # the A matrix is positive definite.  could possibly use conjugate gradient?
  # we made a jax version of this: https://github.com/david-cortes/nonneg_cg
  # but it was still not as fast as the projected gradient solver below.
  # this is possibly due to the line search method, which is a big part
  # of the cost.  jaxopt uses FISTA which we did not implement
  #
  # another avenue worth pursuing is that these A matrices are often
  # fairly sparse.  perhaps worth trying some kind of random or
  # learned projection to solve a smaller dense matrix at each step
  pg = jaxopt.ProjectedGradient(
      objective,
      jaxopt.projection.projection_non_negative,
      maxiter=sys.solver_iterations,
      implicit_diff=False,
      maxls=sys.solver_maxls,
  )

  # solve and convert back to q coordinates
  qf_constraint = state.con_jac.T @ pg.run(jp.zeros_like(b)).params

  return qf_constraint



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wrapper class for facilitating jax transformations on static config data."""

from collections.abc import Iterable
from typing import Any, Dict

from brax.v1 import pytree


@pytree.register
class TracedConfig:
  """A wrapper around the config that faciltates jax tracing of static data."""

  __pytree_ignore__ = ('msg',)

  def __init__(self, msg: Any, custom_tree: Dict[str, Any]):
    self.msg = msg
    self.custom_tree = custom_tree

  def __getattr__(self, name: str):

    base = self.msg.__getattribute__(name)  # pytype:disable=attribute-error

    if name in self.custom_tree:
      if isinstance(base, Iterable) and not isinstance(base, str):
        list_msg = []
        for o, b in zip(self.custom_tree[name], base):
          list_msg.append(TracedConfig(b, custom_tree=o))
        return list_msg
      elif not isinstance(self.custom_tree[name], dict):
        return self.custom_tree[name]
      else:
        return TracedConfig(base, self.custom_tree[name])

    return base



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Composer for environments.

ComponentEnv composes a scene from descriptions of the form below:

   composer = Composer(
    components=dict(
        agent1=dict(component='ant', pos=(0, 1, 0)),
        agent2=dict(component='ant', pos=(0, -1, 0)),
    ),
    edges=dict(ant1__ant2=dict(collide_type='full'),),
   )
   env = ComposeEnv(composer=composer)

 (more examples available in envs/ant_descs.py)

During loading an env through create(), it:
- loads an env_descs, a dictionary containing all args to Composer/ComposerEnv
  - pre-defined envs are defined in envs/
  - new envs can be registered through register_env() or register_lib()
  - example env_descs are in envs/ant_descs.py
  - example multi-agent RL envs (through agent_utils.py) are in envs/ma_descs.py
- creates components: loads and pieces together brax.Config()
    components defined in components/
    such as ant.py or ground.py
  - new components can be registered through register_component()
  - support multiple instances of the same component through comp_name's
  - each component requires: ROOT=root body, SYS_CONFIG=config in string form,
      TERM_FN=termination function of this component, COLLIDES=bodies that
      are allowed to collide, DEFAULT_OBSERVERS=a list of observers (
      see observers.py for references)
  - optionally, each component can specify a dictionary of reward functions
      as `reward_fns`. See reward_functions.py.
- creates edges: automatically create necessary edge information
    among 2+ components, such as collide_include's in brax.Config()
  - optionally edge information can be supplied,
      e.g. `collide_type`={'full', 'root', None} specifying full collisons,
      collision only between roots, or no collision between two components
  - optionally, each edge can specify a dictionary of reward functions
      as `reward_fns`. See reward_functions.py.
- sets reward as sum of all `reward_fns` defined in `components` and `edges`
- sets termination as any(termination_fn of each component)
- sets observation to concatenation of observations of each component defined
    by each component's `observers` argument
"""

import collections
from collections import OrderedDict as odict
import copy
import functools
import itertools
from typing import Dict, Any, Callable, Tuple, Optional, Union

import brax.v1 as brax
from brax.v1 import envs
from brax.v1.envs import Env
from brax.v1.envs import State
from brax.v1.envs import Wrapper
from brax.v1.envs import wrappers
from brax.v1.experimental.braxlines.common import sim_utils
from brax.v1.experimental.braxlines.envs import wrappers as braxlines_wrappers
from brax.v1.experimental.composer import agent_utils
from brax.v1.experimental.composer import component_editor
from brax.v1.experimental.composer import composer_utils
from brax.v1.experimental.composer import data_utils
from brax.v1.experimental.composer import envs as composer_envs
from brax.v1.experimental.composer import observers
from brax.v1.experimental.composer import reward_functions
from brax.v1.experimental.composer.components import load_component
from brax.v1.experimental.composer.components import register_default_components
from jax import numpy as jnp

assert_env_params = composer_envs.assert_env_params
inspect_env = composer_envs.inspect_env
list_env = composer_envs.list_env
register_env = composer_envs.register_env
register_lib = composer_envs.register_lib

register_default_components()
composer_envs.register_default_libs()

MetaData = collections.namedtuple('MetaData', [
    'components',
    'edges',
    'global_options',
    'config_str',
    'config_json',
    'extra_observers',
    'reward_features',
    'reward_fns',
    'agent_groups',
])


class Composer(object):
  """Compose a brax system."""

  def __init__(self,
               components: Dict[str, Dict[str, Any]],
               edges: Dict[str, Dict[str, Any]] = None,
               extra_observers: Tuple[observers.Observer] = (),
               add_ground: bool = True,
               agent_groups: Dict[str, Any] = None,
               global_options: Dict[str, Any] = None):
    components = copy.deepcopy(components)
    edges = copy.deepcopy(edges or {})
    global_options = copy.deepcopy(global_options or {})
    extra_observers = copy.deepcopy(extra_observers)
    reward_features = []
    reward_fns = odict()
    agent_groups = agent_groups or {}

    # load components
    if add_ground:
      components['ground'] = dict(component='ground')
    components = {
        name: load_component(**value) for name, value in components.items()
    }
    component_keys = sorted(components)
    components = odict([(k, components[k]) for k in component_keys])

    # set global
    global_options = dict(
        json=component_editor.json_global_options(**global_options))
    global_options['message_str'] = component_editor.json2message_str(
        global_options['json'])

    components_ = odict()
    for k in component_keys:
      v, new_v = components[k], {}
      # [start config editing] convert to json format for easy editing
      new_v['json'] = component_editor.message_str2json(v.pop('message_str'))

      # add comp_name's
      comp_name = k
      rename_fn = functools.partial(
          component_editor.json_concat_name, comp_name=comp_name)
      new_v['json'] = rename_fn(new_v['json'])
      new_v['collides'] = rename_fn(v.pop('collides'), force_add=True)
      new_v['root'] = rename_fn(v.pop('root'), force_add=True)

      # add frozen bodies option
      if v.pop('frozen', False):
        for b in new_v['json'].get('bodies', []):
          b['frozen'] = {'all': True}
        new_v['frozen'] = True

      # cache config properties
      new_v['bodies'] = [b['name'] for b in new_v['json'].get('bodies', [])]
      new_v['joints'] = [b['name'] for b in new_v['json'].get('joints', [])]
      new_v['actuators'] = [
          b['name'] for b in new_v['json'].get('actuators', [])
      ]
      new_v['comp_name'] = comp_name

      # [end config editing] convert back to str
      new_v['message_str'] = component_editor.json2message_str(new_v['json'])

      # set transform or not
      if 'pos' in v or 'quat' in v:
        new_v['transform'] = True
        new_v['pos'] = jnp.array(v.pop('pos', [0, 0, 0]), dtype='float')
        new_v['quat_origin'] = jnp.array(
            v.pop('quat_origin', [0, 0, 0]), dtype='float')
        new_v['quat'] = jnp.array(
            v.pop('quat', [1., 0., 0., 0.]), dtype='float')
      else:
        new_v['transform'] = False

      # add reward functions
      component_reward_fns = v.pop('reward_fns', {})
      for name, reward_kwargs in sorted(component_reward_fns.items()):
        name = component_editor.concat_name(name, comp_name)
        assert name not in reward_fns, f'duplicate reward_fns {name}'
        reward_fn, unwrapped_reward_fn = reward_functions.get_reward_fns(
            new_v, **reward_kwargs)
        reward_fns[name] = reward_fn
        reward_features += reward_functions.get_observers_from_reward_fns(
            unwrapped_reward_fn)

      # add extra observers
      component_observers = v.pop('extra_observers', ())
      for observer_kwargs in component_observers:
        extra_observers += (observers.get_component_observers(
            new_v, **observer_kwargs),)

      assert all(kk in ('term_fn', 'observers', 'collide')
                 for kk in v), f'unused kwargs in components[{k}]: {v}'
      new_v.update(v)
      components_[k] = new_v
    del components

    edges_ = {}
    for k1, k2 in itertools.combinations(component_keys, 2):
      if k1 == k2:
        continue
      k1, k2 = sorted([k1, k2])  # ensure the name is always sorted in order
      edge_name = component_editor.concat_comps(k1, k2)
      v, new_v = edges.pop(edge_name, {}), {}
      v1, v2 = [components_[k] for k in [k1, k2]]

      # add reward functions
      edge_reward_fns = v.pop('reward_fns', {})
      for name, reward_kwargs in sorted(edge_reward_fns.items()):
        name = component_editor.concat_name(name, edge_name)
        assert name not in reward_fns, f'duplicate reward_fns {name}'
        reward_fn, unwrapped_reward_fn = reward_functions.get_reward_fns(
            v1, v2, **reward_kwargs)
        reward_fns[name] = reward_fn
        reward_features += reward_functions.get_observers_from_reward_fns(
            unwrapped_reward_fn)

      # add observers
      edge_observers = v.pop('extra_observers', ())
      for observer_kwargs in edge_observers:
        extra_observers += (observers.get_edge_observers(
            v1, v2, **observer_kwargs),)

      # [start config editing]
      collide_type = v.pop('collide_type', 'full')
      v_json = {}
      # add colliders
      # TODO: add conaffinity-based collide definitions
      if not all([vv.get('collide', True) for vv in [v1, v2]]):
        pass
      elif collide_type == 'full':
        v_json.update(
            component_editor.json_collides(v1['collides'], v2['collides']))
      elif collide_type == 'root':
        v_json.update(
            component_editor.json_collides([v1['root']], [v2['root']]))
      else:
        assert not collide_type, collide_type
      # [end config editing]
      if v_json:
        # convert back to str
        new_v['message_str'] = component_editor.json2message_str(v_json)
      else:
        new_v['message_str'] = ''
      new_v['json'] = v_json
      assert not v, f'unused kwargs in edges[{edge_name}]: {v}'
      edges_[edge_name] = new_v
    assert not edges, f'unused edges: {edges}'
    edge_keys = sorted(edges_.keys())
    edges_ = odict([(k, edges_[k]) for k in edge_keys])

    # merge all message strs
    message_str = ''
    for _, v in sorted(components_.items()):
      message_str += v.get('message_str', '')
    for _, v in sorted(edges_.items()):
      message_str += v.get('message_str', '')
    message_str += global_options.get('message_str', '')
    config_str = message_str
    config_json = component_editor.message_str2json(message_str)
    metadata = MetaData(
        components=components_,
        edges=edges_,
        global_options=global_options,
        config_str=config_str,
        config_json=config_json,
        extra_observers=extra_observers,
        reward_features=reward_features,
        reward_fns=reward_fns,
        agent_groups=agent_groups,
    )
    config = component_editor.message_str2message(message_str)
    self.config, self.metadata = config, metadata

  def reset_fn(self, sys, qp: brax.QP):
    """Reset state."""
    # apply translations and rotations
    for _, v in sorted(self.metadata.components.items()):
      if v['transform']:
        _, _, mask = sim_utils.names2indices(sys.config, v['bodies'], 'body')
        qp = sim_utils.transform_qp(qp, mask[..., None], v['quat'],
                                    v['quat_origin'], v['pos'])
    return qp

  def term_fn(self, done: jnp.ndarray, sys, qp: brax.QP, info: brax.Info):
    """Termination."""
    for _, v in self.metadata.components.items():
      term_fn = v['term_fn']
      if term_fn:
        done = term_fn(done, sys, qp, info, v)
    return done

  def obs_fn(self, sys, qp: brax.QP, info: brax.Info):
    """Return observation as OrderedDict."""
    cached_obs_dict = {}
    obs_dict = odict()
    reward_features = odict()
    for _, v in self.metadata.components.items():
      for observer in v['observers']:
        obs_dict_ = observers.get_obs_dict(sys, qp, info, observer,
                                           cached_obs_dict, v)
        obs_dict = odict(list(obs_dict.items()) + list(obs_dict_.items()))
    for observer in self.metadata.extra_observers:
      obs_dict_ = observers.get_obs_dict(sys, qp, info, observer,
                                         cached_obs_dict, None)
      obs_dict = odict(list(obs_dict.items()) + list(obs_dict_.items()))
    for observer in self.metadata.reward_features:
      obs_dict_ = observers.get_obs_dict(sys, qp, info, observer,
                                         cached_obs_dict, None)
      reward_features = odict(
          list(reward_features.items()) + list(obs_dict_.items()))
    return obs_dict, reward_features


class ComponentEnv(Env):
  """Make a brax Env fromc config/metadata for training and inference."""

  def __init__(self, composer: Composer, env_desc: Dict[str, Any]):
    self.observer_shapes = None
    self.composer = composer
    self.env_desc = env_desc
    self.metadata = composer.metadata
    super().__init__(config=self.composer.metadata.config_str)
    self.action_shapes = get_action_shapes(self.sys)
    # reward_shape = (num_agents,) if multi-agent else ()
    self.reward_shape = (len(
        self.metadata.agent_groups),) if self.metadata.agent_groups else ()
    assert self.observation_size  # ensure self.observer_shapes is set
    self.group_action_shapes = agent_utils.set_agent_groups(
        self.metadata, self.action_shapes, self.observer_shapes)

  @property
  def is_multiagent(self):
    return bool(self.metadata.agent_groups)

  @property
  def agent_names(self):
    if not self.is_multiagent:
      return ()
    else:
      return tuple(sorted(self.metadata.agent_groups))

  def reset(self, rng: jnp.ndarray) -> State:
    """Resets the environment to an initial state."""
    qpos = self.sys.default_angle()
    qvel = jnp.zeros((self.sys.num_joint_dof,))
    qp = self.sys.default_qp(joint_angle=qpos, joint_velocity=qvel)
    qp = self.composer.reset_fn(self.sys, qp)
    info = self.sys.info(qp)
    obs_dict, _ = self._get_obs(qp, info)
    obs = data_utils.concat_array(obs_dict, self.observer_shapes)
    reward, done, score = jnp.zeros((3,) + self.reward_shape)
    reward_names = tuple(self.composer.metadata.reward_fns)
    if self.is_multiagent:  # multi-agent
      done = jnp.any(done, axis=-1)  # ensure done is a scalar
      reward_names += self.agent_names
    state_info = {}
    state_info['score'] = score
    state_info['rewards'] = odict([(k, jnp.zeros(())) for k in reward_names])
    state_info['scores'] = odict([(k, jnp.zeros(())) for k in reward_names])
    return State(
        qp=qp,
        obs=obs,
        reward=reward,
        done=done.astype(jnp.float32),
        info=state_info)

  def step(self,
           state: State,
           action: jnp.ndarray,
           normalizer_params: Dict[str, jnp.ndarray] = None,
           extra_params: Dict[str, Dict[str, jnp.ndarray]] = None) -> State:
    """Run one timestep of the environment's dynamics."""
    del normalizer_params, extra_params
    qp, info = self.sys.step(state.qp, action)
    obs_dict, reward_features = self._get_obs(qp, info)
    obs = data_utils.concat_array(obs_dict, self.observer_shapes)
    reward_tuple_dict = odict([
        (k, fn(action, reward_features))
        for k, fn in self.composer.metadata.reward_fns.items()
    ])
    state.info['rewards'] = odict([
        (k, v[0]) for k, v in reward_tuple_dict.items()
    ])
    state.info['scores'] = odict([
        (k, v[1]) for k, v in reward_tuple_dict.items()
    ])
    if self.is_multiagent:  # multi-agent
      reward, score, done = agent_utils.process_agent_rewards(
          self.metadata, reward_tuple_dict)
      state.info['rewards'] = odict(
          list(state.info['rewards'].items()) +
          [(k, v) for k, v in zip(self.agent_names, reward)])
      state.info['scores'] = odict(
          list(state.info['scores'].items()) +
          [(k, v) for k, v in zip(self.agent_names, score)])
    else:
      reward, done, score = jnp.zeros((3,))
      for r, s, d in reward_tuple_dict.values():
        reward += r
        score += s
        done = jnp.logical_or(done, d)
    done = self.composer.term_fn(done, self.sys, qp, info)
    state.info['score'] = score
    return state.replace(
        qp=qp, obs=obs, reward=reward, done=done.astype(jnp.float32))

  def _get_obs(self, qp: brax.QP, info: brax.Info) -> jnp.ndarray:
    """Observe."""
    obs_dict, reward_features = self.composer.obs_fn(self.sys, qp, info)
    if self.observer_shapes is None:
      self.observer_shapes = data_utils.get_array_shapes(
          obs_dict, batch_shape=())
    return obs_dict, reward_features


def get_action_shapes(sys):
  """Get action shapes."""
  actuator_names = sim_utils.get_names(sys.config, 'actuator')
  actuator_action_shapes = sim_utils.names2indices(
      sys.config, names=actuator_names, datatype='actuator')[1]
  force_names = sim_utils.get_names(sys.config, 'force')
  force_action_shapes = sim_utils.names2indices(
      sys.config, names=force_names, datatype='force')[1]
  action_shapes = odict([
      (k, dict(start=v[0], end=v[-1] + 1, size=len(v), shape=(len(v),)))
      for k, v in itertools.chain(actuator_action_shapes.items(), force_action_shapes.items())
  ])
  return action_shapes


def unwrap(env: Env):
  """Unwrap all Env wrappers."""
  while isinstance(env, Wrapper):  # unwrap wrappers
    env = env.unwrapped
  return env


def is_multiagent(env: Env):
  """If it supports multiagent RL."""
  env = unwrap(env)
  return env.is_multiagent if hasattr(env, 'is_multiagent') else False


def get_obs_dict_shape(env: Env):
  """Get observation (dictionary) shape."""
  env = unwrap(env)
  if isinstance(env, ComponentEnv):
    assert env.observation_size  # ensure env.observer_shapes is set
    return env.observer_shapes
  else:
    return (env.observation_size,)


def create(env_name: str = None,
           env_desc: Union[Dict[str, Any], Callable[..., Dict[str,
                                                              Any]]] = None,
           desc_edits: Dict[str, Any] = None,
           episode_length: int = 1000,
           action_repeat: int = 1,
           auto_reset: bool = True,
           batch_size: Optional[int] = None,
           **kwargs) -> Env:
  """Creates an Env with a specified brax system."""
  assert env_name or env_desc, 'env_name or env_desc must be supplied'
  env_desc = env_desc or {}
  desc_edits = desc_edits or {}
  if env_name in composer_envs.ENV_DESCS:
    desc = composer_envs.ENV_DESCS[env_name]
    if callable(desc):
      desc = desc(**kwargs)
    else:
      assert not kwargs, f'unused kwargs: {kwargs}'
    env_desc = dict(**env_desc, **desc)
    env_desc = composer_utils.edit_desc(env_desc, desc_edits)
    composer = Composer(**env_desc)
    env = ComponentEnv(composer=composer, env_desc=env_desc)
  elif env_desc:
    if callable(env_desc):
      env_desc = env_desc(**kwargs)
    else:
      assert not kwargs, f'unused kwargs: {kwargs}'
    env_desc = composer_utils.edit_desc(env_desc, desc_edits)
    composer = Composer(**env_desc)
    env = ComponentEnv(composer=composer, env_desc=env_desc)
  else:
    env = envs.create(
        env_name,
        episode_length=episode_length,
        action_repeat=action_repeat,
        auto_reset=auto_reset,
        batch_size=batch_size,
        **kwargs)
    return env  # type: ignore

  # add wrappers
  env = braxlines_wrappers.ExtraStepArgsWrapper(env)
  if episode_length is not None:
    env = wrappers.EpisodeWrapper(env, episode_length, action_repeat)
  if batch_size:
    env = wrappers.VectorWrapper(env, batch_size)
  if auto_reset:
    env = wrappers.AutoResetWrapper(env)
  return env  # type: ignore


def create_fn(env_name: str = None,
              env_desc: Union[Dict[str, Any], Callable[..., Dict[str,
                                                                 Any]]] = None,
              desc_edits: Dict[str, Any] = None,
              episode_length: int = 1000,
              action_repeat: int = 1,
              auto_reset: bool = True,
              batch_size: Optional[int] = None,
              **kwargs) -> Callable[..., Env]:
  """Returns a function that when called, creates an Env."""
  return functools.partial(
      create,
      env_name=env_name,
      env_desc=env_desc,
      desc_edits=desc_edits,
      episode_length=episode_length,
      action_repeat=action_repeat,
      auto_reset=auto_reset,
      batch_size=batch_size,
      **kwargs)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Tests for fluid."""

from absl.testing import absltest
from absl.testing import parameterized
from brax import test_utils
from brax.generalized import dynamics
from brax.generalized import pipeline as g_pipeline
from brax.positional import pipeline as p_pipeline
from brax.spring import pipeline as s_pipeline
import jax
from jax import numpy as jp
import mujoco
import numpy as np

assert_almost_equal = np.testing.assert_almost_equal


class FluidTest(parameterized.TestCase):

  @parameterized.parameters(
      ('fluid_box.xml', 0.0, 0.0),
      ('fluid_box.xml', 0.0, 1.3),
      ('fluid_box.xml', 2.0, 0.0),
      ('fluid_box.xml', 2.0, 1.3),
      ('fluid_box_offset_com.xml', 0.0, 1.3),
      ('fluid_box_offset_com.xml', 2.0, 0.0),
      ('fluid_box_offset_com.xml', 2.0, 1.3),
  )
  def test_fluid_mj_generalized(self, config, density, viscosity):
    """Tests fluid interactions."""
    sys = test_utils.load_fixture(config)
    sys = sys.replace(density=density, viscosity=viscosity)

    mj_model = test_utils.load_fixture_mujoco(config)
    mj_model.opt.density = density
    mj_model.opt.viscosity = viscosity
    mj_data = mujoco.MjData(mj_model)
    # initialize qd so that the object interacts with the fluid
    mj_data.qvel[:3] = [1, 2, 4]
    mj_data.qvel[3] = jp.pi
    q, qd = jp.asarray(mj_data.qpos), jp.asarray(mj_data.qvel)

    # check qfrc_passive after the first step
    mujoco.mj_step(mj_model, mj_data)
    state = jax.jit(g_pipeline.init)(sys, q, qd)
    qfrc_passive = jax.jit(dynamics._passive)(sys, state)
    np.testing.assert_array_almost_equal(
        qfrc_passive[3:], mj_data.qfrc_passive[3:], 2
    )
    m = max(jp.abs(mj_data.qfrc_passive[:3])) + 1e-6
    np.testing.assert_array_almost_equal(
        qfrc_passive[:3] / m, mj_data.qfrc_passive[:3] / m, 1
    )

    # check q/qd after multiple steps
    for _ in range(500):
      mujoco.mj_step(mj_model, mj_data)
    mq, mqd = jp.asarray(mj_data.qpos), jp.asarray(mj_data.qvel)

    for _ in range(500):
      state = jax.jit(g_pipeline.step)(sys, state, jp.zeros((sys.act_size(),)))
    gq, gqd = state.q, state.qd

    np.testing.assert_array_almost_equal(gq, mq, 2)
    np.testing.assert_array_almost_equal(gqd, mqd, 2)

  @parameterized.parameters(
      ('fluid_sphere.xml', p_pipeline, 0.0, 0.0),
      ('fluid_sphere.xml', p_pipeline, 0.0, 1.3),
      ('fluid_sphere.xml', p_pipeline, 2.0, 0.0),
      ('fluid_sphere.xml', p_pipeline, 2.0, 1.3),
      ('fluid_sphere.xml', s_pipeline, 0.0, 0.0),
      ('fluid_sphere.xml', s_pipeline, 0.0, 1.3),
      ('fluid_sphere.xml', s_pipeline, 2.0, 0.0),
      ('fluid_sphere.xml', s_pipeline, 2.0, 1.3),
      ('fluid_two_spheres.xml', p_pipeline, 2.0, 1.3),
      ('fluid_two_spheres.xml', s_pipeline, 2.0, 1.3),
  )
  def test_fluid_positional_spring(self, config, pipeline, density, viscosity):
    """Tests fluid interactions for pbd/spring compared to generalized."""
    sys = test_utils.load_fixture(config)
    sys = sys.replace(density=density, viscosity=viscosity)

    q, qd = sys.init_q, jp.zeros(sys.qd_size())
    qd = qd.at[:3].set(jp.array([1, 2, 4]))
    qd = qd.at[3].set(jp.pi)

    state_g = jax.jit(g_pipeline.init)(sys, q, qd)
    for _ in range(500):
      state_g = jax.jit(g_pipeline.step)(
          sys, state_g, jp.zeros((sys.act_size(),))
      )
    gq, gqd = state_g.q, state_g.qd

    state = jax.jit(pipeline.init)(sys, q, qd)
    for _ in range(500):
      state = jax.jit(pipeline.step)(sys, state, jp.zeros((sys.act_size(),)))
    tq, tqd = state.q, state.qd

    np.testing.assert_array_almost_equal(tq, gq, 3)
    np.testing.assert_array_almost_equal(tqd, gqd, 2)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mujoco XML model loader for Brax."""

import dataclasses
from typing import Any, AnyStr, Dict, List, Optional

from absl import logging
from brax.v1.physics import config_pb2
from dm_control import mjcf
from dm_control.mjcf import constants
import numpy as np
from transforms3d import euler
from transforms3d import quaternions
from transforms3d import utils as transform_utils

Vector = Any
Vector3d = Any
Quaternion = Any

# TODO: In Mujoco, the inertial of a body may be specified using an
# <inertial> element. We currently use a unit diagonal matrix instead.
DEFAULT_INERTIA = [1.0, 1.0, 1.0]
DEFAULT_STIFFNESS = 5000

# Tell the typechecker where to actually look for generic Element properties.
MjcfElement = mjcf.element._ElementImpl


@dataclasses.dataclass
class GeomCollider:
  """Represent a collider for a Mujoco geometry."""
  collider: config_pb2.Collider
  # Volume of the geometry in m^3.
  volume: float = 0
  # Mass of the geometry in kg.
  mass: float = 0


@dataclasses.dataclass
class Collision:
  """Collision information for a body."""
  # See http://www.mujoco.org/book/computation.html#Collision
  contype: int
  conaffinity: int
  # Name of the parent body.
  parent_name: str


def _is_worldbody(mujoco_body: MjcfElement) -> bool:
  """Returns true if the Mujoco body is the worldbody."""
  return mujoco_body.tag == constants.WORLDBODY


def _vec(v: Vector3d) -> config_pb2.Vector3:
  """Converts (numpy) array to Vector3."""
  x, y, z = v
  return config_pb2.Vector3(x=x, y=y, z=z)


def _np_vec(v: config_pb2.Vector3) -> Vector3d:
  """Converts a Vector3 to a numpy array."""
  return np.array([v.x, v.y, v.z])


def _euler(q: Optional[Quaternion]) -> config_pb2.Vector3:
  """Converts the quaternion to Euler angles in degrees."""
  if q is None:
    q = quaternions.qeye()
  return _vec(np.degrees(euler.quat2euler(q, axes='rxyz')))


def _maybe_qmult(q1: Optional[Quaternion],
                 q2: Optional[Quaternion]) -> Quaternion:
  """Returns the multiplication of two quaternions."""
  if q1 is not None:
    return quaternions.qmult(q1, q2) if q2 is not None else q1
  return q2


def _rot_quat(u: Vector3d, v: Vector3d) -> Quaternion:
  """Returns the quaternion performing rotation from u to v."""
  dot_p = np.dot(v, u)
  axis = np.cross(v, u)
  if not np.any(axis):
    return quaternions.qeye()
  norm_axis = np.linalg.norm(axis)
  angle = np.arctan2(norm_axis, dot_p)
  return quaternions.axangle2quat(axis, angle)


def _create_joint(
    name: str,
    parent_body_name: str,
    child_body_name: str,
    parent_offset: Optional[Vector3d] = None,
    child_offset: Optional[Vector3d] = None,
    axis: Optional[Vector3d] = None,
    stiffness: float = DEFAULT_STIFFNESS,
    joint_range: Optional[Vector] = None,
    rotation: Optional[Quaternion] = None,
    reference_rotation: Optional[Quaternion] = None) -> config_pb2.Joint:
  """Returns a (revolute) joint with the specified properties."""
  if axis is None:
    # Default axis of rotation.
    axis = [0, 0, 1]
  if rotation is not None:
    axis = quaternions.rotate_vector(axis, rotation)
  axis = transform_utils.normalized_vector(axis)
  angle_limit = (
      config_pb2.Joint.Range(min=joint_range[0], max=joint_range[1])
      if joint_range is not None else config_pb2.Joint.Range())
  rotation = _rot_quat(axis, [1.0, 0, 0])
  return config_pb2.Joint(
      name=name,
      parent=parent_body_name,
      child=child_body_name,
      parent_offset=_vec(parent_offset) if parent_offset is not None else None,
      child_offset=_vec(child_offset) if child_offset is not None else None,
      stiffness=stiffness,
      rotation=_euler(rotation),
      angle_limit=[angle_limit],
      reference_rotation=_euler(reference_rotation))


def _create_fixed_joint(name: str,
                        parent_body_name: str,
                        child_body_name: str,
                        parent_offset: Optional[Vector3d] = None,
                        rotation: Optional[Quaternion] = None,
                        reference_rotation: Optional[Quaternion] = None):
  """Returns a fixed joint."""
  # Brax does not support such joints. Instead we use a revolute joint with a
  # high stiffness and zero angle range.
  return _create_joint(
      name,
      parent_body_name,
      child_body_name,
      stiffness=DEFAULT_STIFFNESS,
      parent_offset=parent_offset,
      rotation=rotation,
      reference_rotation=reference_rotation)


class MujocoConverter(object):
  """Converts a Mujoco model to a Brax config."""

  def __init__(self,
               xml_string: AnyStr,
               add_collision_pairs: bool = False,
               ignore_unsupported_joints: bool = False,
               add_joint_to_nearest_body: bool = False):
    """Creates a MujocoConverter.

    Args:
      xml_string: XML string containing an Mujoco model description.
      add_collision_pairs: If true, then the collision pairs between bodies will
        be added automatically based on the structure of the model and Mujoco
        collision mask settings. See
        http://www.mujoco.org/book/computation.html#Collision.
      ignore_unsupported_joints: If true, then unsupported joints, e.g. slide,
        will be ignored, otherwise they raise an exception.
      add_joint_to_nearest_body: Adds a joint to the nearest (child)body when
        multiple geometries of a Mujoco body are represented as separate bodies.
    """
    mjcf_model = mjcf.from_xml_string(xml_string, escape_separators=True)
    self._mjcf_model = mjcf_model
    config = config_pb2.Config()
    self._config = config
    self._ignore_unsupported_joints = ignore_unsupported_joints
    self._add_joint_to_nearest_body = add_joint_to_nearest_body
    # Brax uses local coordinates. If global coordinates are used in the
    # Mujoco model, we convert them to local ones.
    self._uses_global = mjcf_model.compiler.coordinate == 'global'
    self._uses_radian = mjcf_model.compiler.angle == 'radian'
    default = mjcf_model.default
    geom = default.geom
    # See http://www.mujoco.org/book/XMLreference.html#geom. Mujoco uses SI
    # units, i.e. m(eter) for size, kg for mass and kg/m^3 for density.
    self._default_density = (
        geom.density if geom.density is not None else 1000.0)
    self._default_contype = geom.contype if geom.contype is not None else 1
    self._default_conaffinity = (
        geom.conaffinity if geom.conaffinity is not None else 1)
    joint = default.joint
    self._default_stiffness = (
        joint.stiffness if joint.stiffness is not None else DEFAULT_STIFFNESS)
    if joint.damping is not None:
      self._config.velocity_damping = joint.damping
    option = mjcf_model.option
    self._config.gravity.CopyFrom(
        _vec(option.gravity if option.gravity is not None else [0, 0, -9.81]))
    self._collisions: Dict[str, Collision] = {}

    # Worldbody is the root of the scene tree. We add the bodies starting from
    # the world body in a depth-first manner.
    self._add_body(mjcf_model.worldbody, None)
    # Add the actuators and the collision pairs.
    self._add_actuators()
    if add_collision_pairs:
      self._add_collision_pairs()

  @property
  def config(self) -> config_pb2.Config:
    """Returns the Brax config for the Mujoco model."""
    return self._config

  def _maybe_to_local(self, pos: Vector3d,
                      mujoco_body: MjcfElement) -> Vector3d:
    """Converts position to local coordinates."""
    if self._uses_global and mujoco_body and not _is_worldbody(mujoco_body):
      return pos - mujoco_body.pos
    return pos

  def _get_position(self, elem: MjcfElement) -> Vector3d:
    """Returns the local position of the Mujoco element, a geom or joint."""
    if elem.pos is None:
      return np.zeros(3)
    return self._maybe_to_local(elem.pos, elem.parent)

  def _maybe_to_radian(self, a: float) -> float:
    """Converts the angle to radian."""
    return a if self._uses_radian else np.radians(a)

  def _get_rotation(self, elem: MjcfElement, axes: str = 'rxyz') -> Quaternion:
    """Returns the rotation quaternion of the Mujoco element."""
    if _is_worldbody(elem):
      return quaternions.qeye()
    if elem.euler is not None:
      return euler.euler2quat(
          self._maybe_to_radian(elem.euler[0]),
          self._maybe_to_radian(elem.euler[1]),
          self._maybe_to_radian(elem.euler[2]),
          axes=axes)
    if elem.quat is not None:
      return np.array([elem.quat[0], elem.quat[1], elem.quat[2], elem.quat[3]])
    if elem.axisangle is not None:
      axisangle = elem.axisangle
      return quaternions.axangle2quat(axisangle[0:3],
                                      self._maybe_to_radian(axisangle[3]))
    return quaternions.qeye()

  def _get_mass(self, geom: MjcfElement, volume: float) -> float:
    """Returns the mass of the geometry based on its volume."""
    if geom.mass:
      return geom.mass
    density = geom.density if geom.density else self._default_density
    return volume * density

  def _get_contype(self, geom: MjcfElement) -> int:
    """Returns the contype of the geometry."""
    return geom.contype if geom.contype is not None else self._default_contype

  def _get_conaffinity(self, geom: MjcfElement) -> int:
    """Returns the conaffinity of the geometry."""
    return (geom.conaffinity
            if geom.conaffinity is not None else self._default_conaffinity)

  def _convert_capsule(self,
                       geom: MjcfElement,
                       add_radius: bool = True) -> GeomCollider:
    """Converts a capsule geometry to a collider."""
    size = geom.size
    radius = size[0]
    if geom.fromto is not None:
      start = geom.fromto[0:3]
      end = geom.fromto[3:6]
      direction = end - start
      length = transform_utils.vector_norm(direction)
      rotation = _rot_quat(direction, [0, 0, 1.0])
      position = self._maybe_to_local((start + end) / 2.0, geom.parent)
    else:
      if len(size) != 2:
        raise ValueError(f'Length is missing for {geom}')
      length = size[1] * 2
      position = self._get_position(geom)
      rotation = quaternions.qeye()
    rotation = _maybe_qmult(rotation, self._get_rotation(geom))
    if add_radius:
      length += 2 * radius
    volume = (np.pi * radius * radius * (length - 2.0 * radius / 3.0))
    mass = self._get_mass(geom, volume)
    return GeomCollider(
        collider=config_pb2.Collider(
            capsule=config_pb2.Collider.Capsule(radius=radius, length=length),
            rotation=_euler(rotation),
            position=_vec(position)),
        volume=volume,
        mass=mass)

  def _convert_cylinder(self, geom: MjcfElement) -> GeomCollider:
    """Converts a cylinder geometry to a collider."""
    return self._convert_capsule(geom, add_radius=False)

  def _convert_sphere(self, geom: MjcfElement) -> GeomCollider:
    """Converts a sphere geometry to a collider."""
    radius = geom.size[0]
    position = self._get_position(geom)
    volume = 4.0 * np.pi * np.power(radius, 3) / 3.0
    mass = self._get_mass(geom, volume)
    return GeomCollider(
        collider=config_pb2.Collider(
            sphere=config_pb2.Collider.Sphere(radius=radius),
            position=_vec(position)),
        volume=volume,
        mass=mass)

  def _convert_box(self, geom: MjcfElement) -> GeomCollider:
    """Converts a box geometry to a collider."""
    position = self._get_position(geom)
    rotation = self._get_rotation(geom)
    size = geom.size
    volume = 8.0 * np.prod(size)
    mass = self._get_mass(geom, volume)
    return GeomCollider(
        collider=config_pb2.Collider(
            box=config_pb2.Collider.Box(halfsize=_vec(size)),
            rotation=_euler(rotation),
            position=position),
        volume=volume,
        mass=mass)

  def _geom_to_collider(self, geom: MjcfElement) -> GeomCollider:
    """Converts a geometry to a collider."""
    if geom.type == 'capsule':
      geom_collider = self._convert_capsule(geom)
    elif geom.type == 'cylinder':
      geom_collider = self._convert_cylinder(geom)
    elif geom.type == 'sphere':
      geom_collider = self._convert_sphere(geom)
    elif geom.type == 'box':
      geom_collider = self._convert_box(geom)
    elif geom.type == 'plane':
      geom_collider = GeomCollider(
          collider=config_pb2.Collider(plane=config_pb2.Collider.Plane()))
    else:
      raise ValueError(f'Unsupported geom {geom.type}.')
    return geom_collider

  def _add_body(self, mujoco_body: MjcfElement,
                parent_body: Optional[config_pb2.Body]):
    """Adds the body, its children bodies and joints to the config."""
    config = self._config
    body_idx = len(config.bodies)
    body = config.bodies.add()
    if not parent_body:
      body.name = constants.WORLDBODY
      body.frozen.position.x, body.frozen.position.y, body.frozen.position.z = (
          1, 1, 1)
      body.frozen.rotation.x, body.frozen.rotation.y, body.frozen.rotation.z = (
          1, 1, 1)
    else:
      body.name = mujoco_body.name if mujoco_body.name else f'$body{body_idx}'
    logging.info('Body %s', body.name)
    geoms = mujoco_body.geom if hasattr(mujoco_body, 'geom') else []
    if not geoms:
      # Except worldbody, we expect the bodies to have a geometry. We add a
      # dummy one if that is not the case.
      geom = mujoco_body.add('geom', pos=[0, 0, 0], type='sphere', size=[0.01])
    else:
      # We add the first geometry to the body itself.
      geom = geoms[0]
    if geom.name and body.name == constants.WORLDBODY:
      # Using the name of the geometry is less confusing for worldbody. It won't
      # use referred to in joints or actuators.
      body.name = geom.name
    reference_rotation = self._get_rotation(mujoco_body)
    geom_collider = self._geom_to_collider(geom)
    body.mass = geom_collider.mass
    body.inertia.CopyFrom(_vec(DEFAULT_INERTIA))
    body.colliders.append(geom_collider.collider)
    geom_colliders = [geom_collider]
    self._collisions[body.name] = Collision(
        contype=self._get_contype(geom),
        conaffinity=self._get_conaffinity(geom),
        parent_name=parent_body.name if parent_body else '')
    # The remaining geometries are added as extra colliders
    for idx, geom in enumerate(geoms[1:]):
      geom_name = geom.name if geom.name else f'${body.name}.{idx}'
      child_geom_collider = self._geom_to_collider(geom)
      body.colliders.append(child_geom_collider.collider)
      body.mass += child_geom_collider.mass
      geom_colliders.append(child_geom_collider)

    if not parent_body:
      # Worldbody cannot have joints.
      pass
    elif mujoco_body.joint:
      self._add_joints(mujoco_body, parent_body, geom_colliders,
                       config.bodies[body_idx:])
    else:
      # If there are no joints between the parent and child body, then this
      # means that they are fixed to each other.
      name = f'${parent_body.name}.{body.name}'
      parent_offset = mujoco_body.pos
      rotation = self._get_rotation(mujoco_body)
      config.joints.append(
          _create_fixed_joint(
              name,
              parent_body.name,
              body.name,
              parent_offset=parent_offset,
              rotation=rotation,
              reference_rotation=reference_rotation))

    for child_mujoco_body in mujoco_body.body:
      self._add_body(child_mujoco_body, body)

  def _add_joints(self, mujoco_body: MjcfElement, parent_body: config_pb2.Body,
                  geom_colliders: List[GeomCollider],
                  geom_bodies: List[config_pb2.Body]):
    """Adds the joints of the body to the config."""
    reference_rotation = self._get_rotation(mujoco_body)
    local_mujoco_body_pos = self._get_position(mujoco_body)
    for joint in mujoco_body.joint:
      if joint.type == 'free':
        continue
      if joint.type != 'hinge':
        if self._ignore_unsupported_joints:
          return
        raise ValueError(
            f'Unsupported joint type {joint.type} for {joint.name}')
      min_i = 0
      if self._add_joint_to_nearest_body:
        # When a body has multiple geometries, all except the first one to are
        # moved to auxiliary child bodies (above), therefore we need to find the
        # one that the joint is actually connected with. For now, we use the
        # distance as a proxy.
        min_d = None
        for i, geom_collider in enumerate(geom_colliders):
          d = np.linalg.norm(self._get_position(joint) -
                             _np_vec(geom_collider.collider.position))
          if min_d is None or min_d > d:
            min_i, min_d = i, d
            logging.info('Distance to %s is %g.', geom_bodies[min_i].name,
                         min_d)
      attachment_body = geom_bodies[min_i]
      local_joint_pos = self._maybe_to_local(joint.pos, mujoco_body)
      logging.info('Joint %s connects %s to %s.', joint.name, parent_body.name,
                   attachment_body.name)
      # The joint position is relative to the Mujoco body. To find the offset
      # with respect to the parent, we need to add the joint and Mujoco body
      # position (i.e. the relative position with respect to the reference point
      # of the parent) and then change the frame of reference to that of the COM
      # of the parent.
      parent_offset = local_mujoco_body_pos + local_joint_pos
      # The attachment geometry will preserve the locality, so it is sufficient
      # to change the frame of reference to its COM.
      child_offset = local_joint_pos
      stiffness = (
          joint.stiffness
          if joint.stiffness is not None else self._default_stiffness)
      self._config.joints.append(
          _create_joint(
              joint.name,
              parent_body.name,
              attachment_body.name,
              parent_offset,
              child_offset,
              axis=joint.axis,
              stiffness=stiffness,
              joint_range=joint.range,
              reference_rotation=reference_rotation))

  def _add_actuators(self):
    """Adds the actuators to the config."""
    for motor in self._mjcf_model.actuator.motor:
      joint_name = motor.joint.name
      self._config.actuators.append(
          config_pb2.Actuator(
              name=joint_name,
              joint=joint_name,
              strength=motor.gear.item(),
              angle=config_pb2.Actuator.Angle()))

  def _add_collision_pairs(self):
    """Adds body pairs that can collide with each other to the config."""
    num_bodies = len(self._config.bodies)
    for i in range(0, num_bodies - 1):
      body1 = self._config.bodies[i]
      col1 = self._collisions[body1.name]
      for j in range(i, num_bodies):
        body2 = self._config.bodies[j]
        col2 = self._collisions[body2.name]
        # See http://www.mujoco.org/book/computation.html#Collision for more
        # information about collision detection in Mujoco.
        if col1.parent_name == col2.parent_name:
          # Two bodies with the same parent cannot collide.
          continue
        if ((col1.parent_name == body2.name and col2.parent_name) or
            (col2.parent_name == body1.name and col1.parent_name)):
          # Parent and child bodies cannot collide unless parent is a geometry
          # of the worldbody.
          continue
        if ((col1.contype & col2.conaffinity) or
            (col2.contype & col1.conaffinity)):
          # Bodies should be compatible for collision.
          ci = self._config.collide_include.add()
          ci.first = body1.name
          ci.second = body2.name
